from copy import copy, deepcopy
from pprint import pformat as prettyformat
from functools import partial

import dask.array
import matplotlib.path
import numpy as np
from scipy.interpolate import griddata as scipy_griddata

import xarray as xr
from xarray import register_dataarray_accessor

from .geometries import apply_geometry
from .plotting.animate import animate_poloidal, animate_pcolormesh, animate_line
from .plotting import plotfuncs
from .plotting.utils import _create_norm
from .region import _from_region
from .utils import (
    _add_cartesian_coordinates,
    _update_metadata_increased_resolution,
    _get_bounding_surfaces,
)


@register_dataarray_accessor("bout")
class BoutDataArrayAccessor:
    """
    Contains BOUT-specific methods to use on BOUT++ dataarrays opened by
    selecting a variable from a BOUT++ dataset.

    These BOUT-specific methods and attributes are accessed via the bout
    accessor, e.g. ``da.bout.options`` returns a `BoutOptionsFile` instance.
    """

    def __init__(self, da):
        self.data = da
        self.metadata = da.attrs.get("metadata")  # None if just grid file
        self.options = da.attrs.get("options")  # None if no inp file

    def __str__(self):
        """
        String representation of the BoutDataArray.

        Accessed by print(da.bout)
        """

        styled = partial(prettyformat, indent=4, compact=True)
        text = (
            "<xbout.BoutDataset>\n"
            + "Contains:\n{}\n".format(str(self.data))
            + "Metadata:\n{}\n".format(styled(self.metadata))
        )
        if self.options:
            text += "Options:\n{}".format(styled(self.options))
        return text

    def to_dataset(self):
        """
        Convert a DataArray to a Dataset, copying the attributes from the DataArray to
        the Dataset, and dropping attributes that only make sense for a DataArray
        """
        da = self.data
        ds = da.to_dataset()

        ds.attrs = deepcopy(da.attrs)

        def dropIfExists(ds, name):
            if name in ds.attrs:
                del ds.attrs[name]

        dropIfExists(ds, "direction_y")
        dropIfExists(ds, "direction_z")
        dropIfExists(ds, "cell_location")

        return ds

    def _shift_z(self, zShift):
        """
        Shift a DataArray in the periodic, toroidal direction using FFTs.

        Parameters
        ----------
        zShift : DataArray
            The angle to shift by
        """
        # Would be nice to use the xrft package for this, but xrft does not currently
        # implement inverse Fourier transforms (although there is an open PR
        # https://github.com/xgcm/xrft/pull/81 to add this).

        # Use dask.array.fft if self.data.data is a dask array
        if isinstance(self.data.data, dask.array.Array):
            fft = dask.array.fft
        else:
            fft = np.fft

        nz = self.data.metadata["nz"]
        # Assume dz is constant here - using FFTs doesn't make much sense if z isn't a
        # toroidal angle coordinate.
        zlength = nz * self.data["dz"].values.flatten()[0]
        nmodes = nz // 2 + 1

        # Get axis position of dimension to transform
        axis = self.data.dims.index(self.data.metadata["bout_zdim"])

        # Create list the dimensions of FFT'd array
        fft_dims = list(deepcopy(self.data.dims))
        fft_dims[axis] = "kz"

        # Fourier transform to get the DataArray in k-space
        data_fft = fft.rfft(self.data.data, axis=axis)

        # Complex phase for rotation by angle zShift
        kz = 2.0 * np.pi * xr.DataArray(np.arange(0, nmodes), dims="kz") / zlength
        phase = 1.0j * kz * zShift

        # Ensure dimensions are in correct order for numpy broadcasting
        extra_dims = deepcopy(fft_dims)
        for dim in phase.dims:
            extra_dims.remove(dim)
        phase = phase.expand_dims(extra_dims)
        phase = phase.transpose(*fft_dims, transpose_coords=True)

        data_shifted_fft = data_fft * np.exp(phase.data)

        data_shifted = fft.irfft(data_shifted_fft, n=nz, axis=axis)

        # Return a DataArray with the same attributes as self, but values from
        # data_shifted
        return self.data.copy(data=data_shifted)

    def to_field_aligned(self):
        """
        Transform DataArray to field-aligned coordinates, which are shifted with respect
        to the base coordinates by an angle zShift
        """
        if (
            self.data.cell_location == "CELL_CENTRE"
            or self.data.cell_location == "CELL_ZLOW"
        ):
            zShift_coord = "zShift"
        else:
            zShift_coord = "zShift_" + self.data.cell_location

        if self.data.direction_y != "Standard":
            raise ValueError(
                f"Cannot shift a {self.data.direction_y} type field to "
                "field-aligned coordinates"
            )

        if zShift_coord not in self.data.coords:
            raise ValueError(
                f"{zShift_coord} missing, cannot shift "
                f"{self.data.cell_location} field {self.data.name} to "
                f"field-aligned coordinates. Setting toroidal geometry is necessary to "
                f'use to_field_aligned() - did you pass the `geometry="toroidal"` '
                f"argument to open_boutdataset()?"
            )

        result = self._shift_z(self.data[zShift_coord])
        result.attrs["direction_y"] = "Aligned"
        return result

    def from_field_aligned(self):
        """
        Transform DataArray from field-aligned coordinates, which are shifted with
        respect to the base coordinates by an angle zShift
        """
        if (
            self.data.cell_location == "CELL_CENTRE"
            or self.data.cell_location == "CELL_ZLOW"
        ):
            zShift_coord = "zShift"
        else:
            zShift_coord = "zShift_" + self.data.cell_location

        if self.data.direction_y != "Aligned":
            raise ValueError(
                f"Cannot shift a {self.data.direction_y} type field from "
                "field-aligned coordinates"
            )

        if zShift_coord not in self.data.coords:
            raise ValueError(
                f"{zShift_coord} missing, cannot shift "
                f"{self.data.cell_location} field {self.data.name} from "
                f"field-aligned coordinates. Setting toroidal geometry is necessary to "
                f'use from_field_aligned() - did you pass the `geometry="toroidal"` '
                f"argument to open_boutdataset()?"
            )

        result = self._shift_z(-self.data[zShift_coord])
        result.attrs["direction_y"] = "Standard"
        return result

    @property
    def _regions(self):
        if "regions" not in self.data.attrs:
            raise ValueError(
                "Called a method requiring regions, but these have not been created. "
                "Please set the 'geometry' option when calling open_boutdataset() to "
                "create regions."
            )
        return self.data.attrs["regions"]

    def from_region(self, name, with_guards=None):
        """
        Get a logically-rectangular section of data from a certain region.
        Includes guard cells from neighbouring regions.

        Parameters
        ----------
        name : str
            Region to get data for
        with_guards : int or dict of int, optional
            Number of guard cells to include, by default use MXG and MYG from BOUT++.
            Pass a dict to set different numbers for different coordinates.
        """
        return _from_region(self.data, name, with_guards)

    @property
    def fine_interpolation_factor(self):
        """
        The default factor to increase resolution when doing parallel interpolation
        """
        return self.data.metadata["fine_interpolation_factor"]

    @fine_interpolation_factor.setter
    def fine_interpolation_factor(self, n):
        """
        Set the default factor to increase resolution when doing parallel interpolation.

        Parameters
        -----------
        n : int
            Factor to increase parallel resolution by
        """
        self.data.metadata["fine_interpolation_factor"] = n

    def interpolate_parallel(
        self,
        region=None,
        *,
        n=None,
        toroidal_points=None,
        method="cubic",
        return_dataset=False,
    ):
        """
        Interpolate in the parallel direction to get a higher resolution version of the
        variable.

        Parameters
        ----------
        region : str, optional
            By default, return a result with all regions interpolated separately and then
            combined. If an explicit region argument is passed, then return the variable
            from only that region.
        n : int, optional
            The factor to increase the resolution by. Defaults to the value set by
            BoutDataset.setupParallelInterp(), or 10 if that has not been called.
        toroidal_points : int or sequence of int, optional
            If int, number of toroidal points to output, applies a stride to toroidal
            direction to save memory usage. If sequence of int, the indexes of toroidal
            points for the output.
        method : str, optional
            The interpolation method to use. Options from xarray.DataArray.interp(),
            currently: linear, nearest, zero, slinear, quadratic, cubic. Default is
            'cubic'.
        return_dataset : bool, optional
            If this is set to True, return a Dataset containing this variable as a member
            (by default returns a DataArray). Only used when region=None.

        Returns
        -------
        A new DataArray containing a high-resolution version of the variable. (If
        return_dataset=True, instead returns a Dataset containing the DataArray.)
        """

        if region is None:
            # Call the single-region version of this method for each region, and combine
            # the results together
            parts = [
                self.interpolate_parallel(
                    region, n=n, toroidal_points=toroidal_points, method=method
                ).bout.to_dataset()
                for region in self._regions
            ]

            # 'region' is not the same for all parts, and should not exist in the result,
            # so delete before merging
            for part in parts:
                if "region" in part.attrs:
                    del part.attrs["region"]
                if "region" in part[self.data.name].attrs:
                    del part[self.data.name].attrs["region"]

            result = xr.combine_by_coords(parts, combine_attrs="drop_conflicts")

            if return_dataset:
                return result
            else:
                # Extract the DataArray to return
                result = apply_geometry(result, self.data.geometry)
                return result[self.data.name]

        # Select a particular 'region' and interpolate to higher parallel resolution
        da = self.data
        region = da.bout._regions[region]
        xcoord = da.metadata["bout_xdim"]
        ycoord = da.metadata["bout_ydim"]
        zcoord = da.metadata["bout_zdim"]

        da = da.bout.from_region(region.name, with_guards={xcoord: 0, ycoord: 2})

        if zcoord in da.dims and da.direction_y != "Aligned":
            aligned_input = False
            da = da.bout.to_field_aligned()
        else:
            aligned_input = True

        if n is None:
            n = self.fine_interpolation_factor

        da = da.chunk({ycoord: None})

        ny_fine = n * region.ny
        dy = (region.yupper - region.ylower) / ny_fine

        myg = da.metadata["MYG"]
        if da.metadata["keep_yboundaries"] and region.connection_lower_y is None:
            ybndry_lower = myg
        else:
            ybndry_lower = 0
        if da.metadata["keep_yboundaries"] and region.connection_upper_y is None:
            ybndry_upper = myg
        else:
            ybndry_upper = 0

        y_fine = np.linspace(
            region.ylower - (ybndry_lower - 0.5) * dy,
            region.yupper + (ybndry_upper - 0.5) * dy,
            ny_fine + ybndry_lower + ybndry_upper,
        )

        # This prevents da.interp() from being very slow.
        # Apparently large attrs (i.e. regions) on a coordinate which is passed as an
        # argument to dask.array.map_blocks() slow things down, maybe because coordinates
        # are numpy arrays, not dask arrays?
        # Slow-down was introduced in d062fa9e75c02fbfdd46e5d1104b9b12f034448f when
        # _add_attrs_to_var(updated_ds, ycoord) was added in geometries.py
        da[ycoord].attrs = {}

        da = da.interp(
            {ycoord: y_fine},
            assume_sorted=True,
            method=method,
            kwargs={"fill_value": "extrapolate"},
        )

        da = _update_metadata_increased_resolution(da, n)

        # Modify dy to be consistent with the higher resolution grid
        dy_array = xr.DataArray(
            np.full([da.sizes[xcoord], da.sizes[ycoord]], dy), dims=[xcoord, ycoord]
        )
        da["dy"] = da["dy"].copy(data=dy_array)

        # Remove regions which have incorrect information for the high-resolution grid.
        # New regions will be generated when creating a new Dataset in
        # BoutDataset.getHighParallelResVars
        del da.attrs["regions"]

        if not aligned_input:
            # Want output in non-aligned coordinates
            da = da.bout.from_field_aligned()

        if toroidal_points is not None and zcoord in da.sizes:
            if isinstance(toroidal_points, int):
                nz = len(da[zcoord])
                zstride = (nz + toroidal_points - 1) // toroidal_points
                da = da.isel(**{zcoord: slice(None, None, zstride)})
            else:
                da = da.isel(**{zcoord: toroidal_points})

        return da

    def add_cartesian_coordinates(self):
        """
        Add Cartesian (X,Y,Z) coordinates.

        Returns
        -------
        DataArray with new coordinates added, which are named 'X_cartesian',
        'Y_cartesian', and 'Z_cartesian'
        """
        return _add_cartesian_coordinates(self.data)

    def remove_yboundaries(self, return_dataset=False, remove_extra_upper=False):
        """
        Remove y-boundary points, if present, from the DataArray

        Parameters
        ----------
        return_dataset : bool, default False
            Return the result as a Dataset containing the new DataArray.
        """

        myg = self.data.metadata["MYG"]

        if (
            self.metadata["keep_yboundaries"] == 0 or myg == 0
        ) and not remove_extra_upper:
            # Ensure we do not modify any other references to metadata
            self.data.attrs["metadata"] = deepcopy(self.data.metadata)
            self.data.metadata["keep_yboundaries"] = 0
            # no y-boundary points to remove
            if return_dataset:
                return self.to_dataset()
            else:
                return self.data
        if self.metadata["keep_yboundaries"] == 0:
            myg = 0

        ycoord = self.data.metadata["bout_ydim"]
        parts = []
        for region in self._regions:
            part = self.data.bout.from_region(region, with_guards=0)
            part_region = list(part.bout._regions.values())[0]
            if part_region.connection_lower_y is None:
                part = part.isel({ycoord: slice(myg, None)})
            if part_region.connection_upper_y is None:
                part = part.isel(
                    {ycoord: slice(-myg if not remove_extra_upper else -myg - 1)}
                )
            del part.attrs["regions"]
            parts.append(part.bout.to_dataset())

        result = xr.combine_by_coords(parts)
        # Ensure we do not modify any other references to metadata
        result.attrs = copy(parts[0].attrs)
        result.attrs["metadata"] = deepcopy(self.data.metadata)
        result[self.data.name].attrs["metadata"] = deepcopy(self.data.metadata)

        # result is as if we had not kept y-boundaries when loading
        result.metadata["keep_yboundaries"] = 0
        result[self.data.name].metadata["keep_yboundaries"] = 0

        if remove_extra_upper:
            # modify jyseps*, ny_inner, ny so that sliced variable gets consistent
            # regions
            if result.metadata["jyseps1_2"] == result.metadata["jyseps2_1"]:
                # single-null
                result.metadata["ny"] -= 1
            else:
                # double-null
                result.metadata["ny_inner"] -= 1
                result.metadata["jyseps1_2"] -= 1
                result.metadata["jyseps2_2"] -= 1
                result.metadata["ny"] -= 2

        if return_dataset:
            return result
        else:
            # Extract the DataArray to return
            return result[self.data.name]

    def ddx(self):
        """
        Special method for calculating a derivative in the "bout_xdim"
        direction (radial, x-direction), needed because the 1d coordinate in this
        direction is index number, not coordinate values (because psi can be different
        in core and PFR regions).

        This method uses a second-order accurate central finite difference method to
        calculate the derivative.

        Note values at the boundaries will be NaN - if Dataset was loaded with
        keep_xboundaries=True, these should only ever be in boundary cells.
        """
        da = self.data
        xcoord = da.metadata["bout_xdim"]

        if da.cell_location == "CELL_CENTRE":
            dx = da["dx"]
        elif da.cell_location == "CELL_XLOW":
            dx = da["dx_CELL_XLOW"]
        elif da.cell_location == "CELL_YLOW":
            dx = da["dx_CELL_YLOW"]
        elif da.cell_location == "CELL_ZLOW":
            dx = da["dx_CELL_ZLOW"]
        else:
            raise ValueError(f'Unrecognised cell location "{da.cell_location}"')

        result = (da.shift({xcoord: -1}) - da.shift({xcoord: 1})) / (2.0 * dx)

        result.name = f"d({da.name})/dx"
        if "standard_name" in result.attrs:
            result.attrs["standard_name"] = f"d({result.attrs['standard_name']})/dx"
        if "long_name" in result.attrs:
            result.attrs["long_name"] = f"x-derivative of {result.attrs['long_name']}"
        if "units" in result.attrs:
            result.attrs["units"] = ""

        return result

    def ddy(self, region=None):
        """
        Special method for calculating a derivative in the "bout_ydim"
        direction (parallel, y-direction), needed because we need to (a) do the
        calculation region-by-region to take account of the branch cuts in the
        y-direction and (b) transform to a field-aligned grid to take parallel
        derivatives.

        This method uses a second-order accurate central finite difference
        method to calculate the derivative. It transforms to a globally field-aligned
        grid to calculate the derivative - there is currently no option to use the same
        method as the BOUT++ approach using 'yup' and 'ydown' fields.

        Note values at the boundaries will be NaN - if Dataset was loaded with
        keep_yboundaries=True, these should only ever be in boundary cells.

        Warnings
        --------
        Depending on how parallel boundary conditions were applied in the BOUT++
        PhysicsModel, the y-boundary cells may not contain valid data, in which case the
        result may be incorrect in the grid cell closest to the boundary.

        Parameters
        ----------
        region : str, optional
            By default, return a result with the derivative calculated in all regions
            separately and then combined. If an explicit region argument is passed, then
            return the result from only that region.

        Returns
        -------
        A new DataArray containing the y-derivative of the variable.
        """
        if region is None:
            # Call the single-region version of this method for each region, and combine
            # the results together
            parts = [self.ddy(r).to_dataset() for r in self._regions]

            # 'region' is not the same for all parts, and should not exist in the
            # result, so delete before merging
            for part in parts:
                if "region" in part.attrs:
                    del part.attrs["region"]

            name = self.data.name
            result = xr.combine_by_coords(parts)[f"d({name})/dy"]

            # regions get mixed up during the split and combine_by_coords, so reset them
            result.attrs["regions"] = self._regions

            return result

        da = self.data
        xcoord = da.metadata["bout_xdim"]
        ycoord = da.metadata["bout_ydim"]
        zcoord = da.metadata["bout_zdim"]

        da = self.data.bout.from_region(region, with_guards={xcoord: 0, ycoord: 1})

        if zcoord in da.dims and da.direction_y != "Aligned":
            aligned_input = False
            da = da.bout.to_field_aligned()
        else:
            aligned_input = True

        if da.cell_location == "CELL_CENTRE":
            dy = da["dy"]
        elif da.cell_location == "CELL_XLOW":
            dy = da["dy_CELL_XLOW"]
        elif da.cell_location == "CELL_YLOW":
            dy = da["dy_CELL_YLOW"]
        elif da.cell_location == "CELL_ZLOW":
            dy = da["dy_CELL_ZLOW"]
        else:
            raise ValueError(f'Unrecognised cell location "{da.cell_location}"')

        result = (da.shift({ycoord: -1}) - da.shift({ycoord: 1})) / (2.0 * dy)

        # Remove any y-guard cells
        region_object = da.bout._regions[region]
        if region_object.connection_lower_y is None:
            ylower = None
        else:
            ylower = 1
        if region_object.connection_upper_y is None:
            yupper = None
        else:
            yupper = -1
        result = result.isel({ycoord: slice(ylower, yupper)})

        if not aligned_input:
            # Want output in non-aligned coordinates
            result = result.bout.from_field_aligned()

        if "regions" in result.attrs:
            del result.attrs["regions"]

        result.name = f"d({da.name})/dy"
        if "standard_name" in result.attrs:
            result.attrs["standard_name"] = f"d({result.attrs['standard_name']})/dy"
        if "long_name" in result.attrs:
            result.attrs["long_name"] = f"y-derivative of {result.attrs['long_name']}"
        if "units" in result.attrs:
            result.attrs["units"] = ""

        return result

    def ddz(self):
        """
        Special method for calculating a derivative in the "bout_zdim"
        direction (toroidal, z-direction), needed because xarray's
        differentiate method doesn't have an option to handle a periodic
        dimension (as of xarray-0.17.0).

        This method uses a second-order accurate central finite difference method to
        calculate the derivative.
        """
        da = self.data
        zcoord = da.metadata["bout_zdim"]
        result = (
            da.roll({zcoord: -1}, roll_coords=False)
            - da.roll({zcoord: 1}, roll_coords=False)
        ) / (2.0 * da["dz"])

        result.name = f"d({da.name})/dz"
        if "standard_name" in result.attrs:
            result.attrs["standard_name"] = f"d({result.attrs['standard_name']})/dz"
        if "long_name" in result.attrs:
            result.attrs["long_name"] = f"z-derivative of {result.attrs['long_name']}"
        if "units" in result.attrs:
            result.attrs["units"] = ""

        return result

    def get_bounding_surfaces(self, coords=("R", "Z")):
        """
        Get bounding surfaces.
        Surfaces are returned as arrays of points describing a polygon, assuming the
        third spatial dimension is a symmetry direction.

        Parameters
        ----------
        coords : (str, str), default ("R", "Z")
            Pair of names of coordinates whose values are used to give the positions of
            the points in the result

        Returns
        -------
        result : list of DataArrays
            Each DataArray in the list contains points on a boundary, with size
            (<number of points in the bounding polygon>, 2). Points wind clockwise around
            the outside domain, and anti-clockwise around the inside (if there is an
            inner boundary).
        """
        return _get_bounding_surfaces(self.data, coords)

    def animate2D(
        self,
        animate_over=None,
        x=None,
        y=None,
        animate=True,
        axis_coords=None,
        fps=10,
        save_as=None,
        ax=None,
        poloidal_plot=False,
        logscale=None,
        **kwargs,
    ):
        """
        Plots a color plot which is animated with time over the specified
        coordinate.

        Currently only supports 2D+1 data, which it plots with animatplot's
        wrapping of matplotlib's pcolormesh.

        Parameters
        ----------
        animate_over : str, optional
            Dimension over which to animate, defaults to the time dimension
        x : str, optional
            Dimension to use on the x axis, default is None - then use the first spatial
            dimension of the data
        y : str, optional
            Dimension to use on the y axis, default is None - then use the second spatial
            dimension of the data
        animate : bool, optional
            If set to false, do not create the animation, just return the block or blocks
        axis_coords : None, str, dict
            Coordinates to use for axis labelling.

            - None: Use the dimension coordinate for each axis, if it exists.
            - "index": Use the integer index values.
            - dict: keys are dimension names, values set axis_coords for each axis
              separately. Values can be: None, "index", the name of a 1d variable or
              coordinate (which must have the dimension given by 'key'), or a 1d
              numpy array, dask array or DataArray whose length matches the length of
              the dimension given by 'key'.

            Only affects time coordinate for plots with poloidal_plot=True.
        fps : int, optional
            Frames per second of resulting gif
        save_as : True or str, optional
            If str is passed, save the animation as save_as+'.gif'.
            If True is passed, save the animation with a default name,
            '<variable name>_over_<animate_over>.gif'
        ax : matplotlib.pyplot.axes object, optional
            Axis on which to plot the gif
        poloidal_plot : bool, optional
            Use animate_poloidal to make a plot in R-Z coordinates (input field must be
            (t,x,y))
        logscale : bool or float, optional
            If True, default to a logarithmic color scale instead of a linear one.
            If a non-bool type is passed it is treated as a float used to set the linear
            threshold of a symmetric logarithmic scale as
            linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is
            passed.
        aspect : str or None, optional
            Argument to set_aspect(). Defaults to "equal" for poloidal plots and "auto"
            for others.
        kwargs : dict, optional
            Additional keyword arguments are passed on to the plotting function
            (animatplot.blocks.Pcolormesh).

        Returns
        -------
        animation or blocks
            If animate==True, returns an animatplot.Animation object, otherwise
            returns a list of animatplot.blocks.Pcolormesh instances.
        """

        data = self.data
        variable = data.name
        n_dims = len(data.dims)

        if n_dims == 3:
            vmin = kwargs.pop("vmin") if "vmin" in kwargs else data.min().values
            vmax = kwargs.pop("vmax") if "vmax" in kwargs else data.max().values
            kwargs["norm"] = _create_norm(
                logscale, kwargs.get("norm", None), vmin, vmax
            )

            if poloidal_plot:
                print(
                    "{} data passed has {} dimensions - making poloidal plot with "
                    "animate_poloidal()".format(variable, str(n_dims))
                )
                if x is not None:
                    kwargs["x"] = x
                if y is not None:
                    kwargs["y"] = y
                poloidal_blocks = animate_poloidal(
                    data,
                    animate_over=animate_over,
                    animate=animate,
                    axis_coords=axis_coords,
                    fps=fps,
                    save_as=save_as,
                    ax=ax,
                    **kwargs,
                )
                return poloidal_blocks
            else:
                print(
                    "{} data passed has {} dimensions - will use "
                    "animatplot.blocks.Pcolormesh()".format(variable, str(n_dims))
                )
                pcolormesh_block = animate_pcolormesh(
                    data=data,
                    animate_over=animate_over,
                    x=x,
                    y=y,
                    animate=animate,
                    axis_coords=axis_coords,
                    fps=fps,
                    save_as=save_as,
                    ax=ax,
                    **kwargs,
                )
                return pcolormesh_block
        else:
            raise ValueError(
                "Data passed has an unsupported number of dimensions "
                "({})".format(str(n_dims))
            )

    def animate1D(
        self,
        animate_over=None,
        animate=True,
        axis_coords=None,
        fps=10,
        save_as=None,
        sep_pos=None,
        ax=None,
        **kwargs,
    ):
        """
        Plots a line plot which is animated over time over the specified coordinate.

        Currently only supports 1D+1 data, which it plots with animatplot's wrapping of
        matplotlib's plot.

        Parameters
        ----------
        animate_over : str, optional
            Dimension over which to animate, defaults to the time dimension
        axis_coords : None, str, dict
            Coordinates to use for axis labelling.

            - None: Use the dimension coordinate for each axis, if it exists.
            - "index": Use the integer index values.
            - dict: keys are dimension names, values set axis_coords for each axis
              separately. Values can be: None, "index", the name of a 1d variable or
              coordinate (which must have the dimension given by 'key'), or a 1d
              numpy array, dask array or DataArray whose length matches the length of
              the dimension given by 'key'.
        fps : int, optional
            Frames per second of resulting gif
        save_as : True or str, optional
            If str is passed, save the animation as save_as+'.gif'.
            If True is passed, save the animation with a default name,
            '<variable name>_over_<animate_over>.gif'
        sep_pos : int, optional
            Radial position at which to plot the separatrix
        ax : Axes, optional
            A matplotlib axes instance to plot to. If None, create a new
            figure and axes, and plot to that
        aspect : str or None, optional
            Argument to ``ax.set_aspect()``, defaults to "auto"
        kwargs : dict, optional
            Additional keyword arguments are passed on to the plotting function
            (animatplot.blocks.Line).

        Returns
        -------
        animation or block
            If ``animate==True``, returns an animatplot.Animation object, otherwise
            returns an animatplot.blocks.Line instance.
        """

        data = self.data
        variable = data.name
        n_dims = len(data.dims)

        if n_dims == 2:
            print(
                "{} data passed has {} dimensions - will use "
                "animatplot.blocks.Line()".format(variable, str(n_dims))
            )
            line_block = animate_line(
                data=data,
                animate_over=animate_over,
                axis_coords=axis_coords,
                sep_pos=sep_pos,
                animate=animate,
                fps=fps,
                save_as=save_as,
                ax=ax,
                **kwargs,
            )
            return line_block

    def interpolate_from_unstructured(
        self,
        *,
        fill_value=np.nan,
        structured_output=True,
        unstructured_dim_name="unstructured_dim",
        **kwargs,
    ):
        """Interpolate DataArray onto new grids of some existing coordinates

        Parameters
        ----------
        **kwargs : (str, array)
            Each keyword is the name of a coordinate in the DataArray, the argument is a
            1d array giving the values of that coordinate on the output grid
        fill_value : float, default np.nan
            fill_value passed through to scipy.interpolation.griddata
        structured_output : bool, default True
            If True, treat output coordinates values as a structured grid.
            If False, output coordinate values must all have the same length and are not
            broadcast together.
        unstructured_dim_name : str, default "unstructured_dim"
            Name used for the dimension in the output that replaces the dimensions of
            the interpolated coordinates. Only used if structured_output=False.

        Returns
        -------
        DataArray
            Data interpolated onto a new, structured grid
        """

        da = self.data

        if structured_output:
            new_coords = {
                name: xr.DataArray(values, dims=name) for name, values in kwargs.items()
            }

            coord_arrays = tuple(
                np.meshgrid(*[values for values in kwargs.values()], indexing="ij")
            )

            new_output_dims = [d for d in kwargs]
        else:
            new_coords = {
                name: xr.DataArray(values, dims=unstructured_dim_name)
                for name, values in kwargs.items()
            }

            coord_arrays = tuple(kwargs.values())

            lengths = [len(c) for c in coord_arrays]
            if np.any([x != lengths[0] for x in lengths[1:]]):
                raise ValueError(
                    f"When structured_output=False, all the arrays of output coordinate"
                    f"values must have the same length. Got lengths "
                    f"{dict((name, len(coord)) for name, coord in kwargs.items())}"
                )

            new_output_dims = [unstructured_dim_name]

        # Figure out number of dimensions in the coordinates to be interpolated
        dims = set()
        for coord in kwargs:
            dims = dims.union(da[coord].dims)
        dims = tuple(dims)
        ndim = len(dims)

        # dimensions that are not being interpolated
        remaining_dims = tuple(d for d in da.dims if d not in dims)

        # Select interpolation method
        if ndim <= 2:
            # "cubic" only available for 1d or 2d interpolation
            method = "cubic"
        else:
            method = "linear"

        # extend input coordinates to cover all dims, so we can flatten them
        for coord in kwargs:
            data = da[coord]
            missing_dims = tuple(set(dims) - set(data.dims))
            expand = {dim: da.sizes[dim] for dim in missing_dims}
            expand_positions = tuple(dims.index(d) for d in missing_dims)
            da[coord] = data.expand_dims(expand, axis=expand_positions)

        # scipy.interpolate.griddata requires the axis being interpolated to be the first
        # one, so stack together 'dims', and then transpose so the resulting stacked
        # dimension is the first
        dims_name_list = [d for d in da.dims if d in dims]
        stacked_dim_name = "stacked_" + "_".join(dims_name_list)
        stacked = da.stack({stacked_dim_name: dims_name_list})
        stacked = stacked.transpose(
            *((stacked_dim_name,) + remaining_dims), transpose_coords=True
        )

        result = scipy_griddata(
            tuple(stacked[coord] for coord in kwargs),
            stacked,
            coord_arrays,
            method=method,
            fill_value=fill_value,
        )

        # griddata only sets points outside the 'convex hull' to fill_value
        # Nicer to set all points outside the grid boundaries to fill_value
        ###################################################################
        boundaries = self.get_bounding_surfaces(coords=[c for c in kwargs])
        points = np.stack(coord_arrays, axis=-1)

        # boundaries[0] is the outer boundary
        path = matplotlib.path.Path(boundaries[0], closed=True, readonly=True)
        is_contained = path.contains_points(points.reshape([-1, 2]))
        is_contained = is_contained.reshape(
            coord_arrays[0].shape + (1,) * len(remaining_dims)
        )
        result = np.where(is_contained, result, fill_value)

        # boundaries[1] is the inner boundary if it exists
        if len(boundaries) > 1:
            path = matplotlib.path.Path(boundaries[1], closed=True, readonly=True)
            is_contained = path.contains_points(points.reshape([-1, 2]))
            is_contained = is_contained.reshape(
                coord_arrays[0].shape + (1,) * len(remaining_dims)
            )
            result = np.where(is_contained, fill_value, result)

        if len(boundaries) > 2:
            raise ValueError(f"Found {len(boundaries)} boundaries, expected at most 2")

        # Create DataArray to return, with as much metadata as possible retained
        ########################################################################
        new_coords.update(
            {
                name: array
                for name, array in stacked.coords.items()
                if stacked_dim_name not in array.dims
            }
        )

        result = xr.DataArray(
            result,
            dims=new_output_dims + list(remaining_dims),
            coords=new_coords,
            name=da.name,
            attrs=da.attrs,
        )

        return result

    def interpolate_to_cartesian(self, *args, **kwargs):
        """
        Interpolate the DataArray to a regular Cartesian grid.

        This method is intended to be used to produce data for visualisation, which
        normally does not require double-precision values, so by default the data is
        converted to `numpy.float32`. Pass ``use_float32=False`` to retain the original
        precision.

        Parameters
        ----------
        nX : int (default 300)
            Number of grid points in the X direction
        nY : int (default 300)
            Number of grid points in the Y direction
        nZ : int (default 100)
            Number of grid points in the Z direction
        use_float32 : bool (default True)
            Downgrade precision to `numpy.float32`?
        fill_value : float (default np.nan)
            Value to use for points outside the interpolation domain (passed to
            `scipy.interpolate.RegularGridInterpolator`)

        See Also
        --------
        BoutDataset.interpolate_to_cartesian
        """
        da = self.data
        name = da.name
        ds = da.to_dataset()
        # Dataset needs geometry and metadata attributes, but these are not copied from
        # the DataArray by default
        ds.attrs["geometry"] = da.geometry
        ds.attrs["metadata"] = da.metadata
        return ds.bout.interpolate_to_cartesian(*args, **kwargs)[name]

    # BOUT-specific plotting functionality: methods that plot on a poloidal (R-Z) plane
    def contour(self, ax=None, **kwargs):
        """
        Contour-plot a radial-poloidal slice on the R-Z plane
        """
        return plotfuncs.plot2d_wrapper(self.data, xr.plot.contour, ax=ax, **kwargs)

    def contourf(self, ax=None, **kwargs):
        """
        Filled-contour-plot a radial-poloidal slice on the R-Z plane
        """
        return plotfuncs.plot2d_wrapper(self.data, xr.plot.contourf, ax=ax, **kwargs)

    def pcolormesh(self, ax=None, **kwargs):
        """
        Colour-plot a radial-poloidal slice on the R-Z plane
        """
        return plotfuncs.plot2d_wrapper(self.data, xr.plot.pcolormesh, ax=ax, **kwargs)

    def plot_regions(self, ax=None, **kwargs):
        """
        Plot the regions into which xBOUT splits radial-poloidal arrays to handle
        tokamak topology.
        """
        return plotfuncs.plot_regions(self.data, ax=ax, **kwargs)

    def plot3d(self, ax=None, **kwargs):
        """
        Make a 3d plot

        Warnings
        --------

        3d plotting functionality is still a bit of a work in progress. Bugs are likely, and
        help developing is welcome!

        Parameters
        ----------

        See plotfuncs.plot3d()
        """
        return plotfuncs.plot3d(self.data, **kwargs)



import numpy as np
import random


from xarray import DataArray, Dataset

from xbout.fastoutput import open_fastoutput


class TestFastOutput:
    def test_open_fastoutput(self, tmp_path):
        datapath_list, xinds, yinds, zinds = make_fastoutput_set(tmp_path, 4)

        ds = open_fastoutput(tmp_path.joinpath("BOUT.fast.*.nc"))

        expected_x = list(set(xinds))
        expected_x.sort()
        expected_y = list(set(yinds))
        expected_y.sort()
        expected_z = list(set(zinds))
        expected_z.sort()

        assert ds.sizes["t"] == 3

        assert all(ds["x"] == expected_x)
        assert all(ds["y"] == expected_y)
        assert all(ds["z"] == expected_z)

        for v in ("n", "T"):
            assert v in ds
            assert "t" in ds[v].dims
            assert "x" in ds[v].dims
            assert "y" in ds[v].dims
            assert "z" in ds[v].dims


def make_fastoutput_set(path, n):
    """
    Create a set of FastOutput files/Datasets, with random data and location indices.

    Parameters
    ----------
    path : pathlib.Path
        Directory to save the files to.
    n : int
        Number of output files/Datasets to create

    Returns
    -------
    result_list : list of pathlib.Path or Dataset
        Output files/Datasets in format duplicating files produced by FastOutput
    xinds : list of int
        x-indices where data exists
    yinds : list of int
        y-indices where data exists
    zinds : list of int
        z-indices where data exists
    """
    # Choose a fixed seed to ensure reproducibility
    random.seed(83)

    procinds = random.sample(range(10 * n), n)
    npoints = [random.randint(1, 4) for _ in procinds]
    total_points = sum(npoints)
    xinds = random.sample(range(max(100, total_points)), total_points)
    yinds = random.sample(range(max(100, total_points)), total_points)
    zinds = random.sample(range(max(100, total_points)), total_points)

    # Add some points which share an x-, y- or z-index (but not all three) with other
    # points
    if total_points < 5:
        raise ValueError(f"Not enough points for n={n}. Increase n.")
    xinds[1] = xinds[0]
    yinds[1] = yinds[0]
    zinds[4] = zinds[3]

    result_list = []

    counter = 0
    for i, procind in enumerate(procinds):
        locations = {}
        for _ in range(npoints[i]):
            locations[counter] = (xinds[counter], yinds[counter], zinds[counter])
            counter += 1
        result_list.append(make_fastoutput(path, procind, locations))

    return result_list, xinds, yinds, zinds


def make_fastoutput(path, i, locations):
    """
    Create a single file in the format produced by
    `FastOutput <https://github.com/johnomotani/BoutFastOutput>`_

    Parameters
    ----------
    path : pathlib.Path
        Directory to save the file to.
    i : int
        Processor index of the file.
    locations : dict of {int: tuple of int}
        Keys are the FastOutput indices of the variables to be saved.
        Values are 3-tuples of ints giving x, y, and z indices associated with the key.
    """

    nt = 3

    # Each variable contains unique random noise
    np.random.seed(seed=i)

    ds = Dataset()

    time = DataArray(np.linspace(0.0, 0.5, nt), dims="t")
    ds["time"] = time

    for j, (xind, yind, zind) in locations.items():
        n = DataArray(np.random.randn(nt), dims="t")
        T = DataArray(np.random.randn(nt), dims="t")

        for v in [n, T]:
            v.attrs["ix"] = xind
            v.attrs["iy"] = yind
            v.attrs["iz"] = zind

        ds[f"n{j}"] = n
        ds[f"T{j}"] = T

    filepath = path.joinpath(f"BOUT.fast.{i}.nc")
    ds.to_netcdf(filepath)



from .load import open_boutdataset, collect

from . import geometries
from .geometries import register_geometry, REGISTERED_GEOMETRIES

from .boutdataset import BoutDatasetAccessor
from .boutdataarray import BoutDataArrayAccessor

from .plotting.animate import animate_pcolormesh, animate_poloidal
from .plotting.utils import plot_separatrix

from .fastoutput import open_fastoutput

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)

__all__ = [
    "open_boutdataset",
    "collect",
    "geometries",
    "register_geometry",
    "REGISTERED_GEOMETRIES",
    "BoutDataArrayAccessor",
    "BoutDatasetAccessor",
    "animate_pcolormesh",
    "animate_poloidal",
    "plot_separatrix",
    "open_fastoutput",
]



from xarray import DataArray, open_dataset, merge
from xarray.testing import assert_equal
import pytest
import numpy as np

from xbout.load import open_boutdataset
from xbout.geometries import register_geometry, REGISTERED_GEOMETRIES


@pytest.fixture
def create_example_grid_file(tmp_path_factory):
    """
    Mocks up a set of BOUT-like netCDF files, and return the temporary test
    directory containing them.

    Deletes the temporary directory once that test is done.
    """

    # Create grid dataset
    arr = np.arange(15).reshape(5, 3)
    grid = DataArray(data=arr, name="arr", dims=["x", "y"]).to_dataset()
    grid["dy"] = DataArray(np.ones((5, 3)), dims=["x", "y"])
    grid = grid.set_coords(["dy"])

    # Create temporary directory
    save_dir = tmp_path_factory.mktemp("griddata")

    # Save
    filepath = save_dir.joinpath("grid.nc")
    grid.to_netcdf(filepath, engine="netcdf4")

    return filepath


class TestOpenGrid:
    def test_open_grid(self, create_example_grid_file):
        example_grid = create_example_grid_file
        with pytest.warns(UserWarning):
            result = open_boutdataset(datapath=example_grid)
        result = result.drop_vars(["x", "y"])
        assert_equal(result, open_dataset(example_grid))
        result.close()

    def test_open_grid_extra_dims(self, create_example_grid_file, tmp_path_factory):
        example_grid = open_dataset(create_example_grid_file)

        new_var = DataArray(
            name="new",
            data=[[1, 2], [8, 9], [16, 17], [27, 28], [37, 38]],
            dims=["x", "w"],
        )

        dodgy_grid_directory = tmp_path_factory.mktemp("dodgy_grid")
        dodgy_grid_path = dodgy_grid_directory.joinpath("dodgy_grid.nc")
        merge([example_grid, new_var]).to_netcdf(dodgy_grid_path, engine="netcdf4")

        with pytest.warns(
            UserWarning, match="drop all variables containing " "the dimensions 'w'"
        ):
            result = open_boutdataset(datapath=dodgy_grid_path)
        result = result.drop_vars(["x", "y"])
        assert_equal(result, example_grid)
        result.close()

    def test_open_grid_apply_geometry(self, create_example_grid_file):
        @register_geometry(name="Schwarzschild")
        def add_schwarzschild_coords(ds, coordinates=None):
            ds["event_horizon"] = 4.0
            ds["event_horizon"].attrs = ds.attrs.copy()
            return ds

        example_grid = create_example_grid_file

        result = result = open_boutdataset(
            datapath=example_grid, geometry="Schwarzschild"
        )
        assert_equal(result["event_horizon"], DataArray(4.0))

        # clean up
        del REGISTERED_GEOMETRIES["Schwarzschild"]
        result.close()

    def test_open_grid_chunks(self, create_example_grid_file):
        example_grid = create_example_grid_file
        with pytest.warns(UserWarning):
            result = open_boutdataset(datapath=example_grid, chunks={"x": 4, "y": 5})
        result = result.drop_vars(["x", "y"])
        assert_equal(result, open_dataset(example_grid))
        result.close()

    def test_open_grid_chunks_not_in_grid(self, create_example_grid_file):
        example_grid = create_example_grid_file
        with pytest.warns(UserWarning):
            result = open_boutdataset(
                datapath=example_grid, chunks={"anonexistantdimension": 5}
            )
        result = result.drop_vars(["x", "y"])
        assert_equal(result, open_dataset(example_grid))
        result.close()



from matplotlib import pyplot as plt

from xbout import open_boutdataset


# We do not distribute binary files with xBOUT, so you need to provide your own gridfile,
# for example one created using Hypnotoad.
gridfilepath = "grid.nc"
grid = open_boutdataset(gridfilepath, geometry="toroidal")

grid["psi_poloidal"].bout.contourf()
grid["psi_poloidal"].bout.contour()
grid["psi_poloidal"].bout.pcolormesh()
grid["psi_poloidal"].bout.pcolormesh(shading="gouraud")

grid["psi_poloidal"].bout.regions()

plt.show()



import pytest


configure_is_imported = False


def pytest_configure(config):
    global configure_is_imported
    if not configure_is_imported:
        configure_is_imported = True
        # filter expected warnings
        config.addinivalue_line(
            "filterwarnings",
            "ignore:No geometry type found, no coordinates will be added:UserWarning",
        )
        config.addinivalue_line(
            "filterwarnings",
            "ignore:deallocating CachingFileManager.*, but file is not already closed. "
            "This may indicate a bug.:RuntimeWarning",
        )

        # register additional markers
        config.addinivalue_line(
            "markers",
            "long: long test, or one of many permutations (disabled by default)",
        )
        config.addinivalue_line(
            "markers",
            "flaky: xarray uses this mark, adding to avoid warnings about it not being "
            "defined",
        )
        config.addinivalue_line(
            "markers",
            "network: xarray uses this mark, adding to avoid warnings about it not "
            "being defined",
        )
        config.addinivalue_line(
            "markers",
            "slow: xarray uses this mark, adding to avoid warnings about it not being "
            "defined",
        )


addoption_is_imported = False


def pytest_addoption(parser):
    # Add command line option '--long' for pytest, to be used to enable long tests
    global addoption_is_imported
    if not addoption_is_imported:
        addoption_is_imported = True
        parser.addoption(
            "--long",
            action="store_true",
            default=False,
            help="enable tests marked as 'long'",
        )


collection_is_imported = False


def pytest_collection_modifyitems(config, items):
    global collection_is_imported
    if not collection_is_imported:
        collection_is_imported = True
        if not config.getoption("--long"):
            # --long not given in cli: skip long tests
            print("\n    skipping long tests, pass '--long' to enable")
            skip_long = pytest.mark.skip(reason="need --long option to run")
            for item in items:
                if "long" in item.keywords:
                    item.add_marker(skip_long)



from copy import deepcopy
from itertools import chain
from pathlib import Path
import os

import numpy as np
import xarray as xr


def _set_attrs_on_all_vars(ds, key, attr_data, copy=False):
    ds.attrs[key] = attr_data
    if copy:
        for v in chain(ds.data_vars, ds.coords):
            ds[v].attrs[key] = deepcopy(attr_data)
    else:
        for v in chain(ds.data_vars, ds.coords):
            ds[v].attrs[key] = attr_data
    return ds


def _add_attrs_to_var(ds, varname, copy=False):
    if copy:
        for attr in ["metadata", "options", "geometry", "regions"]:
            if attr in ds.attrs and attr not in ds[varname].attrs:
                ds[varname].attrs[attr] = deepcopy(ds.attrs[attr])
    else:
        for attr in ["metadata", "options", "geometry", "regions"]:
            if attr in ds.attrs and attr not in ds[varname].attrs:
                ds[varname].attrs[attr] = ds.attrs[attr]


def _check_filetype(path):
    if path.suffix == ".nc":
        filetype = "netcdf4"
    elif path.suffix == ".h5netcdf":
        filetype = "h5netcdf"
    else:
        raise IOError("Do not know how to read file extension {}".format(path.suffix))
    return filetype


def _is_path(p):
    try:
        Path(p)
        return True
    except TypeError:
        return False


def _is_dir(p):
    return _is_path(p) and os.path.isdir(p)


def _separate_metadata(ds):
    """
    Extract the metadata (nxpe, myg etc.) from the Dataset.

    Assumes that all scalar variables are metadata, not physical data!
    """

    # Find only the scalar variables
    variables = list(ds.variables)

    # Remove dz from metadata if it's present. Allows treating dz more consistently
    # whether it is scalar or 2d/3d array.
    exclude = ["dz"]

    scalar_vars = [
        var
        for var in variables
        if not any(dim in ["t", "x", "y", "z"] for dim in ds[var].dims)
        and var not in exclude
    ]

    # Save metadata as a dictionary
    metadata_vals = [ds[var].values.item() for var in scalar_vars]
    metadata_vals = [x.decode() if isinstance(x, bytes) else x for x in metadata_vals]
    metadata = dict(zip(scalar_vars, metadata_vals))

    # Add default values for dimensions to metadata. These may be modified later by
    # apply_geometry()
    metadata["bout_tdim"] = "t"
    metadata["bout_xdim"] = "x"
    metadata["bout_ydim"] = "y"
    metadata["bout_zdim"] = "z"

    return ds.drop_vars(scalar_vars), metadata


def _update_metadata_increased_resolution(da, n):
    """
    Update the metadata variables to account for a y-direction resolution increased by a
    factor n.

    Parameters
    ----------
    da : DataArray
        The variable to update
    n : int
        The factor to increase the y-resolution by
    """

    # Take deepcopy to ensure we do not alter metadata of other variables
    da.attrs["metadata"] = deepcopy(da.metadata)

    def update_jyseps(name):
        # If any jyseps<=0, need to leave as is
        if da.metadata[name] > 0:
            da.metadata[name] = n * (da.metadata[name] + 1) - 1

    update_jyseps("jyseps1_1")
    update_jyseps("jyseps2_1")
    update_jyseps("jyseps1_2")
    update_jyseps("jyseps2_2")

    def update_ny(name):
        da.metadata[name] = n * da.metadata[name]

    update_ny("ny")
    update_ny("ny_inner")
    update_ny("MYSUB")

    # Update attrs of coordinates to be consistent with da
    for coord in da.coords:
        da[coord].attrs = {}
        _add_attrs_to_var(da, coord)

    return da


def _1d_coord_from_spacing(spacing, dim, ds=None, *, origin_at=None):
    """
    Create a 1d coordinate varying along the dimension 'dim' from the grid spacing
    'spacing', with the grid points in the centres of the cells with width given by
    'spacing'.

    Parameters
    ----------
    spacing : DataArray or scalar
        The grid spacing. If a DataArray, must include dimension 'dim' and must be
        constant in any other dimensions.
    dim : str
        The dimension to create a coordinate for
    ds : DataArray or Dataset, optional
        If spacing is a scalar, a Dataset or DataArray containing dimension 'dim' must
        be passed to determine the size of the dimension
    origin_at : string, {["lower"], "centre", "upper"}
        Where to put the origin of the coordinate. Can be at lower edge, centre, or
        upper edge of the grid
    """
    if origin_at is None:
        origin_at = "lower"

    if not isinstance(spacing, xr.DataArray):
        # spacing is a scalar
        if ds is None:
            raise ValueError("ds must be passed when spacing is a scalar")
        n = ds.sizes[dim]
        coord_values = (np.arange(n, dtype=float) + 0.5) * spacing
        total_length = n * spacing
    else:
        other_dims = set(spacing.dims) - set([dim])
        if not np.all(spacing.min(dim=other_dims) == spacing.max(dim=other_dims)):
            raise ValueError(
                f"Spacing is not constant in dimensions other than {dim}. Cannot "
                f"create coordinate"
            )

        point_to_use = {
            spacing.metadata["bout_xdim"]: spacing.metadata.get("MXG", 0)
            if spacing.metadata["keep_xboundaries"]
            else 0,
            spacing.metadata["bout_ydim"]: spacing.metadata.get("MYG", 0)
            if spacing.metadata["keep_yboundaries"]
            else 0,
            spacing.metadata["bout_zdim"]: spacing.metadata.get("MZG", 0),
        }

        # make spacing 1d
        spacing = spacing.isel({d: point_to_use[d] for d in other_dims})

        # xarray stores coordinates as numpy (not dask) arrays anyway, so use .values
        # here to evaluate the task-graph (if there is one)
        spacing = spacing.values

        coord_values = spacing.cumsum()
        total_length = coord_values[-1]

        # adjust to cell mid-point positions
        coord_values = coord_values - 0.5 * spacing

    if origin_at == "lower":
        pass
    elif origin_at == "centre":
        coord_values = coord_values - 0.5 * total_length
    elif origin_at == "upper":
        coord_values = coord_values - total_length
    else:
        raise ValueError(f"Unrecognised argument origin_at={origin_at}")

    return xr.Variable(dim, coord_values)


def _add_cartesian_coordinates(ds):
    # Add Cartesian X and Y coordinates if they do not exist already
    # Works on either BoutDataset or BoutDataArray

    if ds.geometry == "toroidal":
        R = ds["R"]
        Z = ds["Z"]
        zeta = ds[ds.metadata["bout_zdim"]]
        if "X_cartesian" not in ds.coords:
            X = R * np.cos(zeta)
            ds = ds.assign_coords(X_cartesian=X)
        if "Y_cartesian" not in ds.coords:
            Y = R * np.sin(zeta)
            ds = ds.assign_coords(Y_cartesian=Y)
        if "Z_cartesian" not in ds.coords:
            zcoord = ds.metadata["bout_zdim"]
            nz = len(ds[zcoord])
            ds = ds.assign_coords(Z_cartesian=Z.expand_dims({zcoord: nz}, axis=-1))
    else:
        raise ValueError(
            f"Adding Cartesian coordinates to geometry={ds.geometry} is not supported"
        )

    return ds


def _check_new_nxpe(ds, nxpe):
    # Check nxpe is valid

    nx = ds.metadata["nx"] - 2 * ds.metadata["MXG"]

    if nx % nxpe != 0:
        raise ValueError(
            f"nxpe={nxpe} must divide total number of points "
            f"nx-2*MXG={ds.metadata['nx'] - 2*ds.metadata['MXG']}"
        )


def _check_new_nype(ds, nype):
    # Check nype is valid

    ny = ds.metadata["ny"]
    mysub = ny // nype

    if ny % nype != 0:
        raise ValueError(
            f"nype={nype} must divide total number of points ny={ds.metadata['ny']}"
        )
    if (
        ds.metadata["jyseps1_1"] >= 0
        and ds.metadata["jyseps1_1"] <= ny
        and (ds.metadata["jyseps1_1"] + 1) % mysub != 0
    ):
        raise ValueError(
            f"mysub={mysub} must divide jyseps1_1+1={ds.metadata['jyseps1_1'] + 1}"
        )
    if (
        ds.metadata["jyseps2_1"] >= 0
        and ds.metadata["jyseps2_1"] <= ny
        and ds.metadata["jyseps1_2"] != ds.metadata["jyseps2_1"]
        and (ds.metadata["jyseps2_1"] + 1) % mysub != 0
    ):
        raise ValueError(
            f"mysub={mysub} must divide jyseps2_1+1={ds.metadata['jyseps2_1'] + 1}"
        )
    if (
        ds.metadata["jyseps1_2"] != ds.metadata["jyseps2_1"]
        and ds.metadata["ny_inner"] % mysub != 0
    ):
        raise ValueError(
            f"mysub={mysub} must divide ny_inner={ds.metadata['ny_inner']}"
        )
    if (
        ds.metadata["jyseps1_2"] >= 0
        and ds.metadata["jyseps1_2"] <= ny
        and ds.metadata["jyseps1_2"] != ds.metadata["jyseps2_1"]
        and (ds.metadata["jyseps1_2"] + 1) % mysub != 0
    ):
        raise ValueError(
            f"mysub={mysub} must divide jyseps1_2+1={ds.metadata['jyseps1_2'] + 1}"
        )
    if (
        ds.metadata["jyseps2_2"] >= 0
        and ds.metadata["jyseps2_2"] < ny
        and (ds.metadata["jyseps2_2"] + 1) % mysub != 0
    ):
        raise ValueError(
            f"mysub={mysub} must divide jyseps2_2+1={ds.metadata['jyseps2_2'] + 1}"
        )


def _pad_x_boundaries(ds):
    # pad Dataset if boundaries are not already present

    mxg = ds.metadata["MXG"]
    xcoord = ds.metadata.get("bout_xdim", "x")

    if not ds.metadata["keep_xboundaries"] and mxg > 0:
        boundary_pad = ds.isel({xcoord: slice(mxg)}).copy(deep=True).load()
        for v in boundary_pad:
            if xcoord in boundary_pad[v].dims:
                boundary_pad[v].values[...] = np.nan
        ds = xr.concat(
            [boundary_pad, ds.load(), boundary_pad], dim=xcoord, data_vars="minimal"
        )

    return ds


def _pad_y_boundaries(ds):
    # pad Dataset if boundaries are not already present

    myg = ds.metadata["MYG"]
    ycoord = ds.metadata.get("bout_ydim", "y")
    has_second_divertor = ds.metadata["jyseps2_1"] != ds.metadata["jyseps1_2"]

    if not ds.metadata["keep_yboundaries"] and myg > 0:
        boundary_pad = ds.isel({ycoord: slice(myg)}).copy(deep=True).load()
        for v in boundary_pad:
            if ycoord in boundary_pad[v].dims:
                boundary_pad[v].values[...] = np.nan
        if not has_second_divertor:
            ds = xr.concat(
                [boundary_pad, ds.load(), boundary_pad], dim=ycoord, data_vars="minimal"
            )
        else:
            # Include second divertor
            ny_inner = ds.metadata["ny_inner"]
            ds = xr.concat(
                [
                    boundary_pad,
                    ds.isel({ycoord: slice(ny_inner)}).load(),
                    boundary_pad,
                    boundary_pad,
                    ds.isel({ycoord: slice(ny_inner, None)}).load(),
                    boundary_pad,
                ],
                dim=ycoord,
                data_vars="minimal",
            )

    return ds


def _split_into_restarts(ds, variables, savepath, nxpe, nype, tind, prefix, overwrite):
    # Rename dimension coordinates to BOUT++ standard names
    ds = ds.copy()
    for d in ["t", "x", "y", "z"]:
        key = f"bout_{d}dim"
        if key in ds.metadata and ds.metadata[key] in ds.dims:
            ds = ds.rename({ds.metadata[key]: d})

    _check_new_nxpe(ds, nxpe)
    _check_new_nype(ds, nype)

    # Create paths for the files. Do this first so we can check if files exist before
    # spending time creating restart Datasets
    paths = []
    for xproc in range(nxpe):
        for yproc in range(nype):
            # Global processor number
            i = yproc * nxpe + xproc

            paths.append(Path(savepath).joinpath(f"{prefix}.{i}.nc"))
            if not overwrite and paths[-1].exists():
                raise ValueError(
                    f"File {paths[-1]} already exists. Pass overwrite=True to overwrite."
                )

    mxg = ds.metadata["MXG"]
    myg = ds.metadata["MYG"]

    ny_inner = ds.metadata["ny_inner"]

    # These metadata variables are created by xBOUT, so should not be saved to restart
    # files
    restart_exclude_metadata_vars = [
        "bout_tdim",
        "bout_xdim",
        "bout_ydim",
        "bout_zdim",
        "fine_interpolation_factor",
        "is_restart",
        "keep_xboundaries",
        "keep_yboundaries",
    ]

    if variables is None:
        if "t" in ds.dims:
            # If variables to be saved were not specified, add all time-evolving
            # variables
            variables = [v for v in ds if "t" in ds[v].dims]
        else:
            # No time dimension in Dataset, so cannot distinguish time-evolving
            # variables: just include all variables
            variables = [v for v in ds]

    # Add extra variables always needed
    for v in [
        "dx",
        "dy",
        "dz",
        "g11",
        "g22",
        "g33",
        "g12",
        "g13",
        "g23",
        "g_11",
        "g_22",
        "g_33",
        "g_12",
        "g_13",
        "g_23",
        "J",
        "zShift",
    ]:
        if v not in variables and v in ds:
            variables.append(v)

    # number of points in the domain on each processor, not including guard or boundary
    # points
    mxsub = (ds.metadata["nx"] - 2 * mxg) // nxpe
    mysub = ds.metadata["ny"] // nype

    # hist_hi represents the number of iterations before the restart. Attempt to
    # reconstruct here
    final_hist_hi = ds.metadata.get("hist_hi", -1)
    if "t" in ds.dims:
        nt = ds.sizes["t"]
        if tind < 0:
            absolute_tind = tind + nt
            if absolute_tind < 0:
                absolute_tind = 0
        elif tind >= nt:
            absolute_tind = nt - 1
        else:
            absolute_tind = tind
        hist_hi = final_hist_hi - (nt - 1 - absolute_tind)
        if hist_hi < 0:
            hist_hi = -1
    else:
        # Either input is a set of restart files, in which case we should just use
        # `hist_hi` if it exists, or the user has already selected a single time-point,
        # and the best guess we have left is the existing `hist_hi` (although it might
        # not be right).
        hist_hi = final_hist_hi

    has_second_divertor = ds.metadata["jyseps2_1"] != ds.metadata["jyseps1_2"]

    if "t" in ds.dims:
        # select desired time-index for the restart files
        ds = ds.isel({"t": tind}).persist()
        tt = ds["t"].values.flatten()[0]
    else:
        if "tt" in ds.metadata:
            # If loaded from restart files, "tt" should be a scalar in metadata
            tt = ds.metadata["tt"]
        else:
            tt = 0.0

    ds = _pad_x_boundaries(ds)
    ds = _pad_y_boundaries(ds)

    restart_datasets = []
    for xproc in range(nxpe):
        xslice = slice(xproc * mxsub, (xproc + 1) * mxsub + 2 * mxg)
        for yproc in range(nype):
            print(f"creating {xproc*nype + yproc + 1}/{nxpe*nype}", end="\r")
            if has_second_divertor and yproc * mysub >= ny_inner:
                yslice = slice(yproc * mysub + 2 * myg, (yproc + 1) * mysub + 4 * myg)
            else:
                yslice = slice(yproc * mysub, (yproc + 1) * mysub + 2 * myg)

            ds_slice = ds.isel({"x": xslice, "y": yslice})

            restart_ds = xr.Dataset()
            for v in variables:
                data_variable = ds_slice[v].variable

                # delete attrs so we don't try to save metadata dict to restart files
                data_variable.attrs = {}

                restart_ds[v] = data_variable
            for v in ds.metadata:
                if v not in restart_exclude_metadata_vars:
                    value = ds.metadata[v]

                    if isinstance(value, str):
                        # Write strings as byte-strings so BOUT++ can read them
                        value = value.encode()
                    elif isinstance(value, int):
                        value = np.intc(value)

                    restart_ds[v] = value

            # These variables need to be altered, because they depend on the number of
            # files and/or the rank of this file.
            restart_ds["MXSUB"] = np.intc(mxsub)
            restart_ds["MYSUB"] = np.intc(mysub)
            restart_ds["NXPE"] = np.intc(nxpe)
            restart_ds["NYPE"] = np.intc(nype)
            restart_ds["PE_XIND"] = np.intc(xproc)
            restart_ds["PE_YIND"] = np.intc(yproc)
            restart_ds["hist_hi"] = np.intc(hist_hi)
            restart_ds["PE_XIND"] = np.intc(xproc)
            restart_ds["PE_YIND"] = np.intc(yproc)
            restart_ds["MYPE"] = np.intc(yproc * nxpe + xproc)

            # tt is the simulation time where the restart happens
            restart_ds["tt"] = tt

            restart_datasets.append(restart_ds)

    return restart_datasets, paths


# Utility methods for _get_bounding_surfaces()
##############################################

_bounding_surface_checks = {}


def _check_upper_y(ds_region, boundary_points, xbndry, ybndry, Rcoord, Zcoord):
    region = list(ds_region.bout._regions.values())[0]
    xcoord = ds_region.metadata["bout_xdim"]
    ycoord = ds_region.metadata["bout_ydim"]

    if region.connection_upper_y is None:
        xinner = (
            xbndry - 1 if region.connection_inner_x is None and xbndry > 0 else None
        )
        xouter = (
            -xbndry - 1 if region.connection_outer_x is None and xbndry > 0 else None
        )

        # Reverse order of points to go clockwise around region
        boundary_slice = {xcoord: slice(xinner, xouter, -1), ycoord: -1 - ybndry}
        new_boundary_points = np.stack(
            [
                ds_region[Rcoord].isel(boundary_slice).values,
                ds_region[Zcoord].isel(boundary_slice).values,
            ],
            axis=-1,
        )

        boundary_points = np.concatenate([boundary_points, new_boundary_points])

        return boundary_points, region.connection_inner_x, "inner_x"
    else:
        return None


_bounding_surface_checks["upper_y"] = _check_upper_y


def _check_inner_x(ds_region, boundary_points, xbndry, ybndry, Rcoord, Zcoord):
    region = list(ds_region.bout._regions.values())[0]
    xcoord = ds_region.metadata["bout_xdim"]
    ycoord = ds_region.metadata["bout_ydim"]

    if region.connection_inner_x is None:
        ylower = (
            ybndry - 1 if region.connection_lower_y is None and ybndry > 0 else None
        )
        yupper = (
            -ybndry - 1 if region.connection_upper_y is None and ybndry > 0 else None
        )

        # Reverse order of points to go clockwise around region
        boundary_slice = {xcoord: xbndry, ycoord: slice(yupper, ylower, -1)}
        new_boundary_points = np.stack(
            [
                ds_region[Rcoord].isel(boundary_slice).values,
                ds_region[Zcoord].isel(boundary_slice).values,
            ],
            axis=-1,
        )

        boundary_points = np.concatenate([boundary_points, new_boundary_points])

        return boundary_points, region.connection_lower_y, "lower_y"
    else:
        return None


_bounding_surface_checks["inner_x"] = _check_inner_x


def _check_lower_y(ds_region, boundary_points, xbndry, ybndry, Rcoord, Zcoord):
    region = list(ds_region.bout._regions.values())[0]
    xcoord = ds_region.metadata["bout_xdim"]
    ycoord = ds_region.metadata["bout_ydim"]

    if region.connection_lower_y is None:
        xinner = xbndry if region.connection_inner_x is None else None
        xouter = -xbndry if region.connection_outer_x is None and xbndry > 0 else None

        boundary_slice = {xcoord: slice(xinner, xouter), ycoord: ybndry}
        new_boundary_points = np.stack(
            [
                ds_region[Rcoord].isel(boundary_slice).values,
                ds_region[Zcoord].isel(boundary_slice).values,
            ],
            axis=-1,
        )

        boundary_points = np.concatenate([boundary_points, new_boundary_points])

        return boundary_points, region.connection_outer_x, "outer_x"
    else:
        return None


_bounding_surface_checks["lower_y"] = _check_lower_y


def _check_outer_x(ds_region, boundary_points, xbndry, ybndry, Rcoord, Zcoord):
    region = list(ds_region.bout._regions.values())[0]
    xcoord = ds_region.metadata["bout_xdim"]
    ycoord = ds_region.metadata["bout_ydim"]

    if region.connection_outer_x is None:
        ylower = ybndry if region.connection_lower_y is None else None
        yupper = -ybndry if region.connection_upper_y is None and ybndry > 0 else None

        boundary_slice = {xcoord: -1 - xbndry, ycoord: slice(ylower, yupper)}
        new_boundary_points = np.stack(
            [
                ds_region[Rcoord].isel(boundary_slice).values,
                ds_region[Zcoord].isel(boundary_slice).values,
            ],
            axis=-1,
        )

        boundary_points = np.concatenate([boundary_points, new_boundary_points])

        return boundary_points, region.connection_upper_y, "upper_y"
    else:
        return None


_bounding_surface_checks["outer_x"] = _check_outer_x


def _follow_boundary(ds, start_region, start_direction, xbndry, ybndry, Rcoord, Zcoord):
    # Define lists of boundaries to cycle through, depending on which direction we
    # entered the region in
    check_order = {
        "upper_y": ["outer_x", "upper_y", "inner_x", "lower_y"],
        "inner_x": ["upper_y", "inner_x", "lower_y", "outer_x"],
        "lower_y": ["inner_x", "lower_y", "outer_x", "upper_y"],
        "outer_x": ["lower_y", "outer_x", "upper_y", "inner_x"],
    }

    this_region = start_region
    direction = start_direction
    visited_regions = []
    boundary_points = np.empty([0, 2])
    while True:
        # Follow around the boundary from start_region

        visited_regions.append(this_region)

        ds_region = ds.bout.from_region(this_region, with_guards=0)

        # Get all boundary points from this region, and decide which region to go to next
        this_region = None

        for boundary in check_order[direction]:
            result = _bounding_surface_checks[boundary](
                ds_region,
                boundary_points,
                xbndry,
                ybndry,
                Rcoord,
                Zcoord,
            )
            if result is not None:
                boundary_points, this_region, direction = result
            else:
                # Only add the boundary now if it's connected to the boundary we are
                # following
                break

        if this_region is None:
            raise ValueError(f"this_region={this_region} has no boundaries")
        elif this_region == start_region:
            break

    return boundary_points, visited_regions


def _get_bounding_surfaces(ds, coords):
    """
    Get boundary surfaces from a Dataset or DataArray

    Makes some assumptions about possible topologies that are true for currently
    supported single-null/double-null tokamaks, but might not be for very general
    topologies:
    - If there are y-boundaries, there is a region with both a lower-y boundary and an
      outer-x boundary.

    Outer boundary goes clockwise, inner boundary (if present) goes anti-clockwise.

    Note the 'boundary points' are the grid points adjacent to the boundary, not the grid
    boundary half way between grid point and boundary point. Boundary cells are not
    always present, so tricky to implement something 'better'.

    Returns
    -------
    tuple of (N,2) float arrays of vertices on the boundaries
    """

    # coords do not have to be R and Z, but usually will be.
    Rcoord, Zcoord = coords

    # Get number of boundary cells
    if ds.metadata["keep_xboundaries"]:
        xbndry = ds.metadata["MXG"]
    else:
        xbndry = 0
    if ds.metadata["keep_yboundaries"]:
        ybndry = ds.metadata["MYG"]
    else:
        ybndry = 0

    # First find the outer boundary
    start_region = None
    for name, region in ds.bout._regions.items():
        if region.connection_lower_y is None and region.connection_outer_x is None:
            start_region = name
            break
    if start_region is None:
        # No y-boundary region found, presumably is core-only simulation. Start on any
        # region with an outer-x boundary
        for name, region in ds.bout._regions.items():
            if region.connection_outer_x is None:
                start_region = name
    if start_region is None:
        raise ValueError(
            "Failed to find bounding surfaces - no region with outer boundary found"
        )

    # First region has outer-x boundary, but we only visit start_region once, so need to
    # add all boundaries in it the first time.
    region = ds.bout._regions[start_region]
    if region.connection_upper_y is None:
        start_direction = "inner_x"
    elif region.connection_inner_x is None and region.connection_lower_y is None:
        start_direction = "lower_y"
    elif region.connection_lower_y is None:
        start_direction = "outer_x"
    else:
        start_direction = "upper_y"

    boundary, checked_regions = _follow_boundary(
        ds,
        start_region,
        start_direction,
        xbndry,
        ybndry,
        Rcoord,
        Zcoord,
    )
    boundaries = [boundary]

    # Look for an inner boundary
    ############################
    remaining_regions = set(ds.bout._regions) - set(checked_regions)
    start_region = None
    if not remaining_regions:
        # Check for separate inner-x boundary on any of the already visited regions.
        # If inner-x boundary was part of the outer boundary (e.g. for SOL-only
        # simulation) then any region without y-boundaries will be visited twice,
        # otherwise any region that has an inner-x boundary and no y-boundary must be on
        # a separate inner boundary
        for r in checked_regions:
            if (
                ds.bout._regions[r].connection_inner_x is None
                and ds.bout._regions[r].connection_lower_y is not None
                and ds.bout._regions[r].connection_upper_y is not None
                and checked_regions.count(r) < 2
            ):
                start_region = r
                break
    else:
        for r in remaining_regions:
            if (
                ds.bout._regions[r].connection_inner_x is None
                and ds.bout._regions[r].connection_lower_y is not None
                and ds.bout._regions[r].connection_upper_y is not None
            ):
                start_region = r
                break

    if start_region is not None:
        # Inner boundary found
        # Inner boundary should have only an inner_x boundary, so set start_direction to
        # check that first
        start_direction = "lower_y"

        boundary, more_checked_regions = _follow_boundary(
            ds, start_region, start_direction, xbndry, ybndry, Rcoord, Zcoord
        )
        boundaries.append(boundary)
        checked_regions += more_checked_regions

        remaining_regions = set(ds.bout._regions) - set(checked_regions)

    # If there are any remaining regions, they should not have any boundaries
    for r in remaining_regions:
        region = ds.bout._regions[r]
        if (
            region.connection_lower_y is None
            or region.connection_outer_x is None
            or region.connection_upper_y is None
            or region.connection_inner_x is None
        ):
            raise ValueError(
                f"Region '{r}' has a boundary, but was missed when getting boundary "
                f"points"
            )

    # Pack the result into a DataArray
    result = [
        xr.DataArray(
            boundary,
            dims=("boundary", "coord"),
            coords={"coord": [Rcoord, Zcoord]},
        )
        for boundary in boundaries
    ]

    return result


def _set_as_coord(ds, name):
    try:
        ds = ds.set_coords(name)
    except ValueError:
        pass
    try:
        ds = ds.set_coords(f"{name}_CELL_XLOW")
    except ValueError:
        pass
    try:
        ds = ds.set_coords(f"{name}_CELL_YLOW")
    except ValueError:
        pass
    try:
        ds = ds.set_coords(f"{name}_CELL_ZLOW")
    except ValueError:
        pass
    return ds



from xcollect.boutdataset import BoutDataset


bd = BoutDataset(datapath=".")

# Multiply density and temperature fluctuations by 2
for var in ["n", "T"]:
    bd.data["var"] = bd.data["var"] * 2

# Downsample the data by interpolating onto a coarser x coordinate vector
new_x = bd.data["x"].slice(None, None, 3)
bd.data = bd.data.interp(x=new_x)

# Save new restart files with a different parallelisation
bd.save_restart(savepath=".", nxpe=4, nype=8)



from pathlib import Path

import pytest

from xarray import concat
from xarray.tests.test_dataset import create_test_data
import xarray.testing as xrt

from natsort import natsorted

from xbout.load import (
    _check_filetype,
    _expand_wildcards,
    _expand_filepaths,
    _arrange_for_concatenation,
    _trim,
    _infer_contains_boundaries,
    open_boutdataset,
    _BOUT_PER_PROC_VARIABLES,
    _BOUT_TIME_DEPENDENT_META_VARS,
)
from xbout.utils import _separate_metadata
from xbout.tests.utils_for_tests import create_bout_ds, METADATA_VARS


def test_check_extensions(tmp_path):
    files_dir = tmp_path.joinpath("data")
    files_dir.mkdir()
    example_nc_file = files_dir.joinpath("example.nc")
    example_nc_file.write_text("content_nc")

    filetype = _check_filetype(example_nc_file)
    assert filetype == "netcdf4"

    example_hdf5_file = files_dir.joinpath("example.h5netcdf")
    example_hdf5_file.write_text("content_hdf5")

    filetype = _check_filetype(example_hdf5_file)
    assert filetype == "h5netcdf"

    example_invalid_file = files_dir.joinpath("example.txt")
    example_hdf5_file.write_text("content_txt")
    with pytest.raises(IOError):
        filetype = _check_filetype(example_invalid_file)


def test_set_fci_coords(create_example_grid_file_fci, create_example_files_fci):
    grid = create_example_grid_file_fci
    data = create_example_files_fci

    ds = open_boutdataset(data, gridfilepath=grid, geometry="fci")
    assert "R" in ds
    assert "Z" in ds


class TestPathHandling:
    def test_glob_expansion_single(self, tmp_path):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        example_file = files_dir.joinpath("example.0.nc")
        example_file.write_text("content")

        path = example_file
        filepaths = _expand_wildcards(path)
        assert filepaths[0] == example_file

        path = files_dir.joinpath("example.*.nc")
        filepaths = _expand_wildcards(path)
        assert filepaths[0] == example_file

    @pytest.mark.parametrize(
        "ii, jj", [(1, 1), (1, 4), (3, 1), (5, 3), (12, 1), (1, 12), (121, 2), (3, 111)]
    )
    def test_glob_expansion_both(self, tmp_path, ii, jj):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        filepaths = []
        for i in range(ii):
            example_run_dir = files_dir.joinpath("run" + str(i))
            example_run_dir.mkdir()
            for j in range(jj):
                example_file = example_run_dir.joinpath("example." + str(j) + ".nc")
                example_file.write_text("content")
                filepaths.append(example_file)
        expected_filepaths = natsorted(filepaths, key=lambda filepath: str(filepath))

        path = files_dir.joinpath("run*/example.*.nc")
        actual_filepaths = _expand_wildcards(path)

        assert actual_filepaths == expected_filepaths

    @pytest.mark.parametrize(
        "ii, jj", [(1, 1), (1, 4), (3, 1), (5, 3), (1, 12), (3, 111)]
    )
    def test_glob_expansion_brackets(self, tmp_path, ii, jj):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        filepaths = []
        for i in range(ii):
            example_run_dir = files_dir.joinpath("run" + str(i))
            example_run_dir.mkdir()
            for j in range(jj):
                example_file = example_run_dir.joinpath("example." + str(j) + ".nc")
                example_file.write_text("content")
                filepaths.append(example_file)
        expected_filepaths = natsorted(filepaths, key=lambda filepath: str(filepath))

        path = files_dir.joinpath("run[1-9]/example.*.nc")
        actual_filepaths = _expand_wildcards(path)

        assert actual_filepaths == expected_filepaths[jj:]

    def test_no_files(self, tmp_path):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()

        with pytest.raises(IOError):
            path = files_dir.joinpath("run*/example.*.nc")
            _expand_filepaths(path)


@pytest.fixture()
def create_filepaths():
    return _create_filepaths


def _create_filepaths(nxpe=1, nype=1, nt=1):
    filepaths = []
    for t in range(nt):
        for i in range(nype):
            for j in range(nxpe):
                file_num = j + nxpe * i
                path = "./run{}".format(str(t)) + "/BOUT.dmp.{}.nc".format(
                    str(file_num)
                )
                filepaths.append(path)

    return filepaths


class TestArrange:
    def test_arrange_single(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=1, nt=1)
        expected_path_grid = [[["./run0/BOUT.dmp.0.nc"]]]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, None, None]

    def test_arrange_along_x(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=1, nt=1)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc", "./run0/BOUT.dmp.1.nc", "./run0/BOUT.dmp.2.nc"]]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, None, "x"]

    def test_arrange_along_y(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=3, nt=1)
        expected_path_grid = [
            [
                ["./run0/BOUT.dmp.0.nc"],
                ["./run0/BOUT.dmp.1.nc"],
                ["./run0/BOUT.dmp.2.nc"],
            ]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=3
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, "y", None]

    def test_arrange_along_t(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=1, nt=3)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc"]],
            [["./run1/BOUT.dmp.0.nc"]],
            [["./run2/BOUT.dmp.0.nc"]],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", None, None]

    def test_arrange_along_xy(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=2, nt=1)
        expected_path_grid = [
            [
                [
                    "./run0/BOUT.dmp.0.nc",
                    "./run0/BOUT.dmp.1.nc",
                    "./run0/BOUT.dmp.2.nc",
                ],
                [
                    "./run0/BOUT.dmp.3.nc",
                    "./run0/BOUT.dmp.4.nc",
                    "./run0/BOUT.dmp.5.nc",
                ],
            ]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=2
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, "y", "x"]

    def test_arrange_along_xt(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=1, nt=2)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc", "./run0/BOUT.dmp.1.nc", "./run0/BOUT.dmp.2.nc"]],
            [["./run1/BOUT.dmp.0.nc", "./run1/BOUT.dmp.1.nc", "./run1/BOUT.dmp.2.nc"]],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", None, "x"]

    def test_arrange_along_xyt(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=2, nt=2)
        expected_path_grid = [
            [
                [
                    "./run0/BOUT.dmp.0.nc",
                    "./run0/BOUT.dmp.1.nc",
                    "./run0/BOUT.dmp.2.nc",
                ],
                [
                    "./run0/BOUT.dmp.3.nc",
                    "./run0/BOUT.dmp.4.nc",
                    "./run0/BOUT.dmp.5.nc",
                ],
            ],
            [
                [
                    "./run1/BOUT.dmp.0.nc",
                    "./run1/BOUT.dmp.1.nc",
                    "./run1/BOUT.dmp.2.nc",
                ],
                [
                    "./run1/BOUT.dmp.3.nc",
                    "./run1/BOUT.dmp.4.nc",
                    "./run1/BOUT.dmp.5.nc",
                ],
            ],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=2
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", "y", "x"]


class TestStripMetadata:
    def test_strip_metadata(self):
        original = create_bout_ds()
        assert original["NXPE"] == 1

        ds, metadata = _separate_metadata(original)

        assert original.drop_vars(
            METADATA_VARS + _BOUT_PER_PROC_VARIABLES + _BOUT_TIME_DEPENDENT_META_VARS,
            errors="ignore",
        ).equals(ds)
        assert metadata["NXPE"] == 1


# TODO also test loading multiple files which have guard cells
class TestOpen:
    def test_single_file(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=1, nype=1, nt=1, write_to_disk=True
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)
        expected = create_bout_ds()
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS
                + _BOUT_PER_PROC_VARIABLES
                + _BOUT_TIME_DEPENDENT_META_VARS,
                errors="ignore",
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(None, nxpe=1, nype=1, nt=1)
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_squashed_file(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=3, nt=1, squashed=True, write_to_disk=True
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)
        expected = create_bout_ds(lengths=(6, 8, 12, 7))
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS
                + _BOUT_PER_PROC_VARIABLES
                + _BOUT_TIME_DEPENDENT_META_VARS,
                errors="ignore",
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(None, nxpe=4, nype=3, nt=1, squashed=True)
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    @pytest.mark.parametrize(
        "keep_xboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    @pytest.mark.parametrize(
        "keep_yboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    def test_squashed_doublenull(
        self,
        tmp_path_factory,
        bout_xyt_example_files,
        keep_xboundaries,
        keep_yboundaries,
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=6,
            nt=1,
            lengths=(6, 2, 4, 7),
            guards={"x": 2, "y": 2},
            squashed=True,
            topology="lower-disconnected-double-null",
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=path,
                keep_xboundaries=keep_xboundaries,
                keep_yboundaries=keep_yboundaries,
            )

        # bout_xyt_example_files when creating a 'squashed' file just makes it with
        # y-size nype*lengths[2]+2*myg, which is 6*4+4=28, so with upper and lower
        # boundaries removed, y-size should be 28-4*2=20.
        assert ds.sizes["t"] == 6
        assert ds.sizes["x"] == 12 if keep_xboundaries else 8
        assert ds.sizes["y"] == 32 if keep_yboundaries else 24
        assert ds.sizes["z"] == 7

    @pytest.mark.parametrize(
        "keep_xboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    @pytest.mark.parametrize(
        "keep_yboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    def test_squashed_doublenull_file(
        self,
        tmp_path_factory,
        bout_xyt_example_files,
        keep_xboundaries,
        keep_yboundaries,
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=6,
            nt=1,
            lengths=(6, 4, 4, 7),
            guards={"x": 2, "y": 2},
            squashed=True,
            write_to_disk=True,
            topology="upper-disconnected-double-null",
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=path,
                keep_xboundaries=keep_xboundaries,
                keep_yboundaries=keep_yboundaries,
            )

        # bout_xyt_example_files when creating a 'squashed' file just makes it with
        # y-size nype*lengths[2]+2*myg, which is 6*4+4=28, so with upper and lower
        # boundaries removed, y-size should be 28-4*2=20.
        assert ds.sizes["t"] == 6
        assert ds.sizes["x"] == 20 if keep_xboundaries else 16
        assert ds.sizes["y"] == 32 if keep_yboundaries else 24
        assert ds.sizes["z"] == 7

    def test_combine_along_x(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=1,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        bout_ds = create_bout_ds
        expected = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2), bout_ds(3)],
            dim="x",
            data_vars="minimal",
        )
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS + _BOUT_PER_PROC_VARIABLES, errors="ignore"
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None, nxpe=4, nype=1, nt=1, syn_data_type="stepped"
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_combine_along_y(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=1,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        bout_ds = create_bout_ds
        expected = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2)], dim="y", data_vars="minimal"
        )
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS + _BOUT_PER_PROC_VARIABLES, errors="ignore"
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None, nxpe=1, nype=3, nt=1, syn_data_type="stepped"
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    @pytest.mark.skip
    def test_combine_along_t(self):
        ...

    @pytest.mark.parametrize(
        "bout_v5,metric_3D", [(False, False), (True, False), (True, True)]
    )
    @pytest.mark.parametrize("lengths", [(6, 2, 4, 7), (6, 2, 4, 1)])
    def test_combine_along_xy(
        self, tmp_path_factory, bout_xyt_example_files, bout_v5, metric_3D, lengths
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=3,
            nt=1,
            lengths=lengths,
            syn_data_type="stepped",
            write_to_disk=True,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        def bout_ds(syn_data_type):
            return create_bout_ds(
                syn_data_type, bout_v5=bout_v5, metric_3D=metric_3D, lengths=lengths
            )

        line1 = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2), bout_ds(3)],
            dim="x",
            data_vars="minimal",
        )
        line2 = concat(
            [bout_ds(4), bout_ds(5), bout_ds(6), bout_ds(7)],
            dim="x",
            data_vars="minimal",
        )
        line3 = concat(
            [bout_ds(8), bout_ds(9), bout_ds(10), bout_ds(11)],
            dim="x",
            data_vars="minimal",
        )
        expected = concat([line1, line2, line3], dim="y", data_vars="minimal")
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        vars_to_drop = METADATA_VARS + _BOUT_PER_PROC_VARIABLES
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(vars_to_drop, errors="ignore"),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None,
            nxpe=4,
            nype=3,
            nt=1,
            lengths=lengths,
            syn_data_type="stepped",
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_toroidal(self, tmp_path_factory, bout_xyt_example_files):
        # actually write these to disk to test the loading fully
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
            write_to_disk=True,
        )
        actual = open_boutdataset(
            datapath=path,
            geometry="toroidal",
            gridfilepath=path.parent.joinpath("grid.nc"),
        )

        # check dataset can be saved
        save_dir = tmp_path_factory.mktemp("data")
        actual.bout.save(save_dir.joinpath("boutdata.nc"))

        # check creation without writing to disk gives identical result
        fake_ds_list, fake_grid_ds = bout_xyt_example_files(
            None,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
        )
        fake = open_boutdataset(
            datapath=fake_ds_list, geometry="toroidal", gridfilepath=fake_grid_ds
        )
        xrt.assert_identical(actual, fake)

    def test_salpha(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
            write_to_disk=True,
        )
        actual = open_boutdataset(
            datapath=path,
            geometry="s-alpha",
            gridfilepath=path.parent.joinpath("grid.nc"),
        )

        # check dataset can be saved
        save_dir = tmp_path_factory.mktemp("data")
        actual.bout.save(save_dir.joinpath("boutdata.nc"))

        # check creation without writing to disk gives identical result
        fake_ds_list, fake_grid_ds = bout_xyt_example_files(
            None, nxpe=3, nype=3, nt=1, syn_data_type="stepped", grid="grid"
        )
        fake = open_boutdataset(
            datapath=fake_ds_list, geometry="s-alpha", gridfilepath=fake_grid_ds
        )
        xrt.assert_identical(actual, fake)

    def test_drop_vars(self, tmp_path_factory, bout_xyt_example_files):
        datapath = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=1,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=datapath, keep_xboundaries=False, drop_variables=["T"]
            )

        assert "T" not in ds.keys()
        assert "n" in ds.keys()

    @pytest.mark.skip
    def test_combine_along_tx(self):
        ...

    def test_restarts(self):
        datapath = Path(__file__).parent.joinpath(
            "data", "restart", "BOUT.restart.*.nc"
        )
        ds = open_boutdataset(datapath, keep_xboundaries=True, keep_yboundaries=True)

        assert "T" in ds


_test_processor_layouts_list = [
    # No parallelization
    (0, 0, 1, 1, {"x": True, "y": True}, {"x": True, "y": True}),
    # 1d parallelization along x:
    # Left
    (0, 0, 3, 1, {"x": True, "y": True}, {"x": False, "y": True}),
    # Middle
    (1, 0, 3, 1, {"x": False, "y": True}, {"x": False, "y": True}),
    # Right
    (2, 0, 3, 1, {"x": False, "y": True}, {"x": True, "y": True}),
    # 1d parallelization along y:
    # Bottom
    (0, 0, 1, 3, {"x": True, "y": True}, {"x": True, "y": False}),
    # Middle
    (0, 1, 1, 3, {"x": True, "y": False}, {"x": True, "y": False}),
    # Top
    (0, 2, 1, 3, {"x": True, "y": False}, {"x": True, "y": True}),
    # 2d parallelization:
    # Bottom left corner
    (0, 0, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    # Bottom right corner
    (2, 0, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    # Top left corner
    (0, 3, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    # Top right corner
    (2, 3, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
    # Centre
    (1, 2, 3, 4, {"x": False, "y": False}, {"x": False, "y": False}),
    # Left side
    (0, 2, 3, 4, {"x": True, "y": False}, {"x": False, "y": False}),
    # Right side
    (2, 2, 3, 4, {"x": False, "y": False}, {"x": True, "y": False}),
    # Bottom side
    (1, 0, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    # Top side
    (1, 3, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
]

_test_processor_layouts_doublenull_list = [
    # 1d parallelization along y:
    # Bottom
    (0, 0, 1, 4, {"x": True, "y": True}, {"x": True, "y": False}),
    # Lower Middle
    (0, 1, 1, 4, {"x": True, "y": False}, {"x": True, "y": True}),
    # Upper Middle
    (0, 2, 1, 4, {"x": True, "y": True}, {"x": True, "y": False}),
    # Top
    (0, 3, 1, 4, {"x": True, "y": False}, {"x": True, "y": True}),
    # 2d parallelization:
    # Bottom left corner
    (0, 0, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    (1, 0, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    # Bottom right corner
    (2, 0, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    (0, 1, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    (1, 1, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
    (2, 1, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
    (0, 2, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    (1, 2, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    (2, 2, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    # Top left corner
    (0, 3, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    (1, 3, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
    # Top right corner
    (2, 3, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
]


class TestTrim:
    @pytest.mark.parametrize("is_restart", [False, True])
    def test_no_trim(self, is_restart):
        ds = create_test_data(0)
        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"
        actual = _trim(
            ds, guards={}, keep_boundaries={}, nxpe=1, nype=1, is_restart=is_restart
        )
        xrt.assert_equal(actual, ds)

    def test_trim_guards(self):
        ds = create_test_data(0)
        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"
        actual = _trim(
            ds, guards={"time": 2}, keep_boundaries={}, nxpe=1, nype=1, is_restart=False
        )
        selection = {"time": slice(2, -2)}
        expected = ds.isel(**selection)
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_list,
    )
    def test_infer_boundaries_2d_parallelization(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        .. code:: text

            y  9 10 11
            ^  6 7  8
            |  3 4  5
            |  0 1  2
             -----> x
        """

        ds = create_test_data(0)
        ds["jyseps2_1"] = 0
        ds["jyseps1_2"] = 0
        ds["PE_XIND"] = xproc
        ds["PE_YIND"] = yproc
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_doublenull_list,
    )
    def test_infer_boundaries_2d_parallelization_doublenull(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        .. code:: text

            y  9 10 11
            ^  6 7  8
            |  3 4  5
            |  0 1  2
             -----> x
        """

        ds = create_test_data(0)
        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4
        ds["PE_XIND"] = xproc
        ds["PE_YIND"] = yproc
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_list,
    )
    def test_infer_boundaries_2d_parallelization_by_filenum(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        .. code:: text

            y  9 10 11
            ^  6 7  8
            |  3 4  5
            |  0 1  2
             -----> x
        """

        filenum = yproc * nxpe + xproc

        ds = create_test_data(0)
        ds["jyseps2_1"] = 0
        ds["jyseps1_2"] = 0
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_doublenull_list,
    )
    def test_infer_boundaries_2d_parallelization_doublenull_by_filenum(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        .. code:: text

            y  9 10 11
            ^  6 7  8
            |  3 4  5
            |  0 1  2
             -----> x
        """

        filenum = yproc * nxpe + xproc

        ds = create_test_data(0)
        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_xboundaries(self, is_restart):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "x"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"

        ds["jyseps2_1"] = 8
        ds["jyseps1_2"] = 8

        actual = _trim(
            ds,
            guards={"x": 2},
            keep_boundaries={"x": True},
            nxpe=1,
            nype=1,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_yboundaries(self, is_restart):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "y"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"

        ds["jyseps2_1"] = 8
        ds["jyseps1_2"] = 8

        actual = _trim(
            ds,
            guards={"y": 2},
            keep_boundaries={"y": True},
            nxpe=1,
            nype=1,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize(
        "filenum, lower, upper",
        [(0, True, False), (1, False, True), (2, True, False), (3, False, True)],
    )
    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_yboundaries_doublenull_by_filenum(
        self, filenum, lower, upper, is_restart
    ):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "y"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"

        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4

        actual = _trim(
            ds,
            guards={"y": 2},
            keep_boundaries={"y": True},
            nxpe=1,
            nype=4,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        if not lower:
            expected = expected.isel(y=slice(2, None, None))
        if not upper:
            expected = expected.isel(y=slice(None, -2, None))
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_trim_timing_info(self, is_restart):
        ds = create_test_data(0)
        from xbout.load import _BOUT_PER_PROC_VARIABLES

        # remove a couple of entries from _BOUT_PER_PROC_VARIABLES so we test that _trim
        # does not fail if not all of them are present
        _BOUT_PER_PROC_VARIABLES = _BOUT_PER_PROC_VARIABLES[:-2]

        for v in _BOUT_PER_PROC_VARIABLES:
            ds[v] = 42.0
        ds = _trim(
            ds, guards={}, keep_boundaries={}, nxpe=1, nype=1, is_restart=is_restart
        )

        expected = create_test_data(0)
        xrt.assert_equal(ds, expected)



import re
import xarray as xr
from xbout.load import _expand_filepaths


def open_fastoutput(datapath="BOUT.fast.*.nc"):
    """
    Opens fast output data and combines into a single dataset.

    """

    # Get list of all files
    filepaths, filetype = _expand_filepaths(datapath)

    # Iterate over all files, extracting DataArrays ready for combining
    fo_data = []
    for i, filepath in enumerate(filepaths):
        fo = xr.open_dataset(filepath)

        if i == 0:
            # Get time coordinate from first file
            time = fo["time"]

        # Time is global, and we already extracted it
        fo = fo.drop_vars("time", errors="ignore")

        # There might be no virtual probe in this region
        if len(fo.data_vars) > 0:
            for name, da in fo.items():
                # Save the physical position (in index units)
                da = da.expand_dims(x=1, y=1, z=1)
                da = da.assign_coords(
                    x=xr.DataArray([da.attrs["ix"]], dims=["x"]),
                    y=xr.DataArray([da.attrs["iy"]], dims=["y"]),
                    z=xr.DataArray([da.attrs["iz"]], dims=["z"]),
                )

                # Re-attach the time coordinate
                da = da.assign_coords(time=time)

                # We saved the position, so don't care what number the variable was
                # Only need it's name (i.e. n, T, etc.)
                regex = re.compile(r"(\D+)([0-9]+)")
                match = regex.match(name)
                if match is None:
                    raise ValueError(f"Regex could not parse the variable named {name}")
                var, num = match.groups()
                da.name = var

                # Must promote DataArrays to Datasets until we require xarray-0.19.0
                # where xarray GH #3248 is fixed
                ds = xr.Dataset({var: da})
                fo_data.append(ds)

        fo.close()

    # This will merge different variables, and arrange by physical position
    full_fo = xr.combine_by_coords(fo_data, combine_attrs="drop_conflicts")

    return full_fo



from warnings import warn
from textwrap import dedent

import xarray as xr
import numpy as np

from .region import _create_regions_toroidal, _create_single_region
from .utils import (
    _add_attrs_to_var,
    _set_attrs_on_all_vars,
    _set_as_coord,
    _1d_coord_from_spacing,
)

REGISTERED_GEOMETRIES = {}


class UnregisteredGeometryError(Exception):
    """Error for unregistered geometry type"""


# TODO remove coordinates argument completely. All the functionality should
# instead be implemented by defining a particular geometry that does what is
# desired


def apply_geometry(ds, geometry_name, *, coordinates=None, grid=None):
    """

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset (from
    geometry_name : str
        Name under which the desired geometry function was registered
    coordinates : dict of str, optional
        Names to give the physical coordinates corresponding to 'x', 'y' and 'z'; values
        corresponding to 'x', 'y' and 'z' keys in the passed dict are used as the names
        of the dimensions. Any not passed are given default values. If not specified,
        default names are chosen.
    grid : Dataset, optional
        Dataset containing extra geometrical information not stored in the dump files
        that is needed to add coordinates for the geometry being applied. For example,
        should contain 2d arrays Rxy, Zxy and psixy for toroidal geometry.

    Returns
    -------
    updated_ds

    Raises
    ------
    UnregisteredGeometryError
    """

    if geometry_name is None or geometry_name == "":
        ds = _set_attrs_on_all_vars(ds, "geometry", "")
        updated_ds = ds
    else:
        try:
            add_geometry_coords = REGISTERED_GEOMETRIES[geometry_name]
        except KeyError:
            message = dedent(
                """{} is not a registered geometry. Inspect the global
                             variable REGISTERED_GEOMETRIES to see which geometries
                             have been registered.""".format(
                    geometry_name
                )
            )
            raise UnregisteredGeometryError(message)

        # User-registered functions may accept 'coordinates' and 'grid' arguments, but
        # do not have to as long as they are not used
        if coordinates is not None and grid is not None:
            updated_ds = add_geometry_coords(ds, coordinates=coordinates, grid=grid)
        elif coordinates is not None:
            updated_ds = add_geometry_coords(ds, coordinates=coordinates)
        elif grid is not None:
            updated_ds = add_geometry_coords(ds, grid=grid)
        else:
            updated_ds = add_geometry_coords(ds)

        # Set "geometry" attribute after adding coordinates, so that functions in
        # `REGISTERED_GEOMETRIES` can check if ds.attrs["geometry"] is already defined
        # to see if they are being applied for the first time or re-applied.
        updated_ds = _set_attrs_on_all_vars(updated_ds, "geometry", geometry_name)

    del ds

    # Set dimension names if they were not set by add_geometry_coords(). Dimensions
    # should not have been renamed without having set their names already.
    if "bout_tdim" not in updated_ds.metadata:
        if "t" in updated_ds.dims:
            updated_ds.metadata["bout_tdim"] = "t"
        else:
            raise ValueError(
                '"t" dimension was renamed, but metadata["bout_tdim"] was not set'
            )
    if "bout_xdim" not in updated_ds.metadata:
        if "x" in updated_ds.dims:
            updated_ds.metadata["bout_xdim"] = "x"
        else:
            raise ValueError(
                '"x" dimension was renamed, but metadata["bout_xdim"] was not set'
            )
    if "bout_ydim" not in updated_ds.metadata:
        if "y" in updated_ds.dims:
            updated_ds.metadata["bout_ydim"] = "y"
        else:
            raise ValueError(
                '"y" dimension was renamed, but metadata["bout_ydim"] was not set'
            )
    if "bout_zdim" not in updated_ds.metadata:
        if "z" in updated_ds.dims:
            updated_ds.metadata["bout_zdim"] = "z"
        else:
            raise ValueError(
                '"z" dimension was renamed, but metadata["bout_zdim"] was not set'
            )

    # Add global 1D coordinates
    # ######################
    # Note the global coordinates used here are defined so that they are zero at
    # the boundaries of the grid (where the grid includes all boundary cells), not
    # necessarily the physical boundaries, because constant offsets do not matter, as
    # long as these bounds are consistent with the global coordinates defined in
    # Region.__init__() (we will only use these coordinates for interpolation) and it is
    # simplest to calculate them with cumsum().
    tcoord = updated_ds.metadata["bout_tdim"]
    xcoord = updated_ds.metadata["bout_xdim"]
    ycoord = updated_ds.metadata["bout_ydim"]
    zcoord = updated_ds.metadata["bout_zdim"]

    if (tcoord not in updated_ds.coords) and (tcoord in updated_ds.dims):
        # Create the time coordinate from t_array
        # Slightly odd looking way to create coordinate ensures 'index variable' is
        # created, which using set_coords() does not (possible xarray bug?
        # https://github.com/pydata/xarray/issues/4417
        updated_ds[tcoord] = updated_ds["t_array"]
        updated_ds = updated_ds.drop_vars("t_array")

    if xcoord not in updated_ds.coords:
        # Make index 'x' a coordinate, useful for handling global indexing
        # Note we have to use the index value, not the value calculated from 'dx' because
        # 'dx' may not be consistent between different regions (e.g. core and PFR).
        # For some geometries xcoord may have already been created by
        # add_geometry_coords, in which case we do not need this.
        nx = updated_ds.dims[xcoord]

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # updated_ds = updated_ds.assign_coords(**{xcoord: np.arange(nx)})
        updated_ds[xcoord] = (xcoord, np.arange(nx))

        _add_attrs_to_var(updated_ds, xcoord)

    if ycoord not in updated_ds.coords:
        # dy should always be constant in x, so it is safe to convert to a 1d
        # coordinate.  [The y-coordinate has to be a 1d coordinate that labels x-z
        # slices of the grid (similarly x-coordinate is 1d coordinate that labels y-z
        # slices and z-coordinate is a 1d coordinate that labels x-y slices). A
        # coordinate might have different values in disconnected regions, but there are
        # no branch-cuts allowed in the x-direction in BOUT++ (at least for the
        # momement), so the y-coordinate has to be 1d and single-valued. Therefore
        # similarly dy has to be 1d and single-valued.]

        # calculate ycoord at the centre of each cell
        y = _1d_coord_from_spacing(updated_ds["dy"], ycoord)

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs as _1d_coord_from_spacing returns an xr.Variable
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # updated_ds = updated_ds.assign_coords(**{ycoord: y.values})
        updated_ds[ycoord] = y

        _add_attrs_to_var(updated_ds, ycoord)

    # If full data (not just grid file) then toroidal dim will be present
    if zcoord in updated_ds.dims and zcoord not in updated_ds.coords:
        # Generates a coordinate whose value is 0 on the first grid point, not dz/2, to
        # match how BOUT++ generates fields from input file expressions.
        nz = updated_ds.dims[zcoord]

        # In BOUT++ v5, dz is either a Field2D or Field3D.
        # We can use it as a 1D coordinate if it's a Field3D, _or_ if nz == 1
        bout_version = updated_ds.metadata.get("BOUT_VERSION", 4.3)
        bout_v5 = bout_version > 5.0 or (
            bout_version == 5.0 and updated_ds["dz"].ndim >= 2
        )
        use_metric_3d = updated_ds.metadata.get("use_metric_3d", False)
        can_use_1d_z_coord = (nz == 1) or use_metric_3d

        if can_use_1d_z_coord:
            if updated_ds.geometry == "fci":
                # dz is varying. just set to a linspace
                z = xr.DataArray(
                    np.linspace(start=0, stop=2 * np.pi, num=nz, endpoint=False),
                    dims=zcoord,
                )
            else:
                z = _1d_coord_from_spacing(updated_ds["dz"], zcoord, updated_ds)
        else:
            if bout_v5:
                if not np.all(updated_ds["dz"].min() == updated_ds["dz"].max()):
                    raise ValueError(
                        "Spacing is not constant. Cannot create z coordinate"
                    )

                dz = updated_ds["dz"].min()
            else:
                dz = updated_ds["dz"]

            z0 = 2 * np.pi * updated_ds.metadata["ZMIN"]
            z1 = z0 + nz * dz
            if not np.all(
                np.isclose(
                    z1,
                    2.0 * np.pi * updated_ds.metadata["ZMAX"],
                    rtol=1.0e-15,
                    atol=0.0,
                )
            ):
                warn(
                    f"Size of toroidal domain as calculated from nz*dz ({str(z1 - z0)}"
                    f" is not the same as 2pi*(ZMAX - ZMIN) ("
                    f"{2.*np.pi*updated_ds.metadata['ZMAX'] - z0}): using value from dz"
                )
            z = xr.DataArray(
                np.linspace(start=z0, stop=z1, num=nz, endpoint=False), dims=zcoord
            )

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # updated_ds = updated_ds.assign_coords(**{zcoord: z})
        updated_ds[zcoord] = (zcoord, z.values)

        _add_attrs_to_var(updated_ds, zcoord)

    # Add dx, dy and dz as coordinates, so that they are available with BoutDataArrays
    updated_ds = _set_as_coord(updated_ds, "dx")
    updated_ds = _set_as_coord(updated_ds, "dy")
    updated_ds = _set_as_coord(updated_ds, "dz")

    return updated_ds


def register_geometry(name):
    """
    Register a new geometry type.

    Used as a decorator on a user-defined function which accepts a dataset and
    returns a new dataset, The returned dataset should have dimensions changed
    and coordinates added as appropriate for this new geometry type.

    Parameters
    ----------
    name: str
        Name under which this geometry should be registered.

    """

    # TODO if anything more needs to be done with geometries than just adding
    # coordinates once, they should be reimplemented as class inheritance

    if not isinstance(name, str):
        raise ValueError("The name of the new geometry type must be a string")

    def wrapper(add_geometry_coords):
        if name in REGISTERED_GEOMETRIES:
            warn(
                "There is aready a geometry type registered with the "
                "name {}, it will be overrided".format(name)
            )

        REGISTERED_GEOMETRIES[name] = add_geometry_coords

        return add_geometry_coords

    return wrapper


def _set_default_toroidal_coordinates(coordinates, ds):
    if coordinates is None:
        coordinates = {}

    # Replace any values that have not been passed in with defaults
    if ds.metadata["is_restart"] == 0:
        # Don't need "t" coordinate for restart files which have no time dimension, and
        # adding it breaks the check for reloading in open_boutdataset
        coordinates["t"] = coordinates.get("t", ds.metadata["bout_tdim"])

    default_x = (
        ds.metadata["bout_xdim"] if ds.metadata["bout_xdim"] != "x" else "psi_poloidal"
    )
    coordinates["x"] = coordinates.get("x", default_x)

    default_y = ds.metadata["bout_ydim"] if ds.metadata["bout_ydim"] != "y" else "theta"
    coordinates["y"] = coordinates.get("y", default_y)

    default_z = ds.metadata["bout_zdim"] if ds.metadata["bout_zdim"] != "z" else "zeta"
    coordinates["z"] = coordinates.get("z", default_z)

    return coordinates


def _add_vars_from_grid(ds, grid, variables, *, optional_variables=None):
    # Get extra geometry information from grid file if it's not in the dump files
    for v in variables:
        if v not in ds:
            if grid is None:
                raise ValueError(
                    f"Grid file is required to provide {v}. Pass the grid "
                    f"file name as the 'gridfilepath' argument to "
                    f"open_boutdataset()."
                )
            # ds[v] = grid[v]
            # Work around issue where xarray drops attributes on coordinates when a new
            # DataArray is assigned to the Dataset, see
            # https://github.com/pydata/xarray/issues/4415
            # https://github.com/pydata/xarray/issues/4393
            # This way adds as a 'Variable' instead of as a 'DataArray'
            ds[v] = (grid[v].dims, grid[v].values)

            _add_attrs_to_var(ds, v)

    if optional_variables is not None:
        for v in optional_variables:
            if v not in ds:
                if grid is None:
                    continue
                if v in grid:
                    # ds[v] = grid[v]
                    # Work around issue where xarray drops attributes on
                    # coordinates when a new DataArray is assigned to the
                    # Dataset, see https://github.com/pydata/xarray/issues/4415
                    # https://github.com/pydata/xarray/issues/4393
                    # This way adds as a 'Variable' instead of as a 'DataArray'
                    ds[v] = (grid[v].dims, grid[v].values)

                    _add_attrs_to_var(ds, v)

    return ds


@register_geometry("toroidal")
def add_toroidal_geometry_coords(ds, *, coordinates=None, grid=None):
    coordinates = _set_default_toroidal_coordinates(coordinates, ds)

    if ds.attrs.get("geometry", None) == "toroidal":
        # Loading a Dataset which already had the coordinates created for it
        ds = _create_regions_toroidal(ds)
        return ds

    # Check whether coordinates names conflict with variables in ds
    bad_names = [
        name for name in coordinates.values() if name in ds and name not in ds.coords
    ]
    if bad_names:
        raise ValueError(
            "Coordinate names {} clash with variables in the dataset. "
            "Register a different geometry to provide alternative names. "
            "It may be useful to use the 'coordinates' argument to "
            "add_toroidal_geometry_coords() for this.".format(bad_names)
        )

    # Get extra geometry information from grid file if it's not in the dump files
    ds = _add_vars_from_grid(
        ds,
        grid,
        ["psixy", "Rxy", "Zxy"],
        optional_variables=[
            "Bpxy",
            "Brxy",
            "Bzxy",
            "poloidal_distance",
            "poloidal_distance_ylow",
            "total_poloidal_distance",
            "zShift",
            "zShift_ylow",
        ],
    )

    if "t" in ds.dims:
        # Rename 't' if user requested it
        ds = ds.rename(t=coordinates["t"])

    # Change names of dimensions to Orthogonal Toroidal ones
    ds = ds.rename(y=coordinates["y"])

    # TODO automatically make this coordinate 1D in simplified cases?
    ds = ds.rename(psixy=coordinates["x"])
    ds = ds.set_coords(coordinates["x"])
    ds[coordinates["x"]].attrs["units"] = "Wb"

    # Record which dimensions 't', 'x', and 'y' were renamed to.
    if ds.metadata["is_restart"] == 0:
        ds.metadata["bout_tdim"] = coordinates["t"]
    # x dimension not renamed, so this is still 'x'
    ds.metadata["bout_xdim"] = "x"
    ds.metadata["bout_ydim"] = coordinates["y"]

    # If full data (not just grid file) then toroidal dim will be present
    if "z" in ds.dims:
        ds = ds.rename(z=coordinates["z"])

        # Record which dimension 'z' was renamed to.
        ds.metadata["bout_zdim"] = coordinates["z"]

    # Ensure metadata is the same on all variables
    ds = _set_attrs_on_all_vars(ds, "metadata", ds.metadata)

    # Add 2D Cylindrical coordinates
    if ("R" not in ds) and ("Z" not in ds):
        ds = ds.rename(Rxy="R", Zxy="Z")
        ds = ds.set_coords(("R", "Z"))
    else:
        ds = ds.set_coords(("Rxy", "Zxy"))

    # Rename zShift_ylow if it was added from grid file, to be consistent with name if
    # it was added from dump file
    if "zShift_CELL_YLOW" in ds and "zShift_ylow" in ds:
        # Remove redundant copy
        del ds["zShift_ylow"]
    elif "zShift_ylow" in ds:
        ds = ds.rename(zShift_ylow="zShift_CELL_YLOW")

    if "poloidal_distance" in ds:
        ds = ds.set_coords(
            ["poloidal_distance", "poloidal_distance_ylow", "total_poloidal_distance"]
        )

    # Add zShift as a coordinate, so that it gets interpolated along with a variable
    ds = _set_as_coord(ds, "zShift")
    if "zShift_CELL_YLOW" in ds:
        ds = _set_as_coord(ds, "zShift_CELL_YLOW")

    ds = _create_regions_toroidal(ds)

    return ds


@register_geometry("s-alpha")
def add_s_alpha_geometry_coords(ds, *, coordinates=None, grid=None):
    coordinates = _set_default_toroidal_coordinates(coordinates, ds)

    if set(coordinates.values()).issubset(set(ds.coords).union(ds.dims)):
        # Loading a Dataset which already had the coordinates created for it
        ds = _create_regions_toroidal(ds)
        return ds

    ds = add_toroidal_geometry_coords(ds, coordinates=coordinates, grid=grid)

    # Get extra geometry information from grid file if it's not in the dump files
    # Add 'hthe' from grid file, needed below for radial coordinate
    if "hthe" not in ds:
        hthe_from_grid = True
        ycoord = "y"
        if grid is None:
            raise ValueError(
                "Grid file is required to provide %s. Pass the grid "
                "file name as the 'gridfilepath' argument to "
                "open_boutdataset()."
            )
        ds["hthe"] = grid["hthe"]
        _add_attrs_to_var(ds, "hthe")
    else:
        hthe_from_grid = False
        ycoord = coordinates["y"]

    # Add 1D radial coordinate
    if "r" in ds:
        raise ValueError(
            "Cannot have variable 'r' in dataset when using " "geometry='s-alpha'"
        )
    ds["r"] = ds["hthe"].isel({ycoord: 0}).squeeze(drop=True)
    ds["r"].attrs["units"] = "m"
    ds = ds.set_coords("r")
    ds = ds.rename(x="r")
    ds.metadata["bout_xdim"] = "r"

    if hthe_from_grid:
        # remove hthe because it does not have correct metadata
        del ds["hthe"]

    return ds


@register_geometry("fci")
def add_fci_geometry_coords(ds, *, coordinates=None, grid=None):
    assert coordinates is None, "Not implemented"
    ds = _add_vars_from_grid(ds, grid, ["R", "Z"])
    ds = ds.set_coords(("R", "Z"))
    ds = _create_single_region(ds, periodic_y=True)
    return ds



import pytest

import numpy as np
import xarray as xr
import xarray.testing as xrt

from xbout.utils import (
    _set_attrs_on_all_vars,
    _update_metadata_increased_resolution,
    _1d_coord_from_spacing,
)


class TestUtils:
    def test__set_attrs_on_all_vars(self):
        ds = xr.Dataset()
        ds["a"] = xr.DataArray([0.0])
        ds["b"] = xr.DataArray([1.0])

        _set_attrs_on_all_vars(ds, "i", 42)
        _set_attrs_on_all_vars(ds, "metadata", {"x": 3})

        assert ds.i == 42
        assert ds["a"].i == 42
        assert ds["b"].i == 42

        ds.metadata["x"] = 5
        # no deepcopy done, so metadata on 'a' and 'b' should have changed too
        assert ds["a"].metadata["x"] == 5
        assert ds["b"].metadata["x"] == 5

    def test__set_attrs_on_all_vars_copy(self):
        ds = xr.Dataset()
        ds["a"] = xr.DataArray([0.0])
        ds["b"] = xr.DataArray([1.0])

        _set_attrs_on_all_vars(ds, "metadata", {"x": 3}, copy=True)

        ds.metadata["x"] = 5
        # deepcopy done, so metadata on 'a' and 'b' should not have changed
        assert ds.metadata["x"] == 5
        assert ds["a"].metadata["x"] == 3
        assert ds["b"].metadata["x"] == 3

    def test__update_metadata_increased_resolution(self):
        da = xr.DataArray()
        da.attrs["metadata"] = {
            "jyseps1_1": 1,
            "jyseps2_1": 2,
            "ny_inner": 3,
            "jyseps1_2": 4,
            "jyseps2_2": 5,
            "ny": 6,
            "MYSUB": 7,
        }

        da = _update_metadata_increased_resolution(da, 3)

        assert da.metadata["jyseps1_1"] == 5
        assert da.metadata["jyseps2_1"] == 8
        assert da.metadata["jyseps1_2"] == 14
        assert da.metadata["jyseps2_2"] == 17

        assert da.metadata["ny_inner"] == 9
        assert da.metadata["ny"] == 18
        assert da.metadata["MYSUB"] == 21

    @pytest.mark.parametrize(
        "origin_at", [None, "lower", "centre", "upper", "expectfail"]
    )
    def test__1d_coord_from_spacing_scalar(self, origin_at):
        ds = xr.Dataset()
        ds["foo"] = ("testdim", np.arange(10))

        if origin_at == "expectfail":
            # fail with no ds for scalar argument
            with pytest.raises(ValueError):
                coord = _1d_coord_from_spacing(0.2, "testdim")
            with pytest.raises(ValueError):
                coord = _1d_coord_from_spacing(0.2, "testdim", ds, origin_at=origin_at)
            return

        coord = _1d_coord_from_spacing(0.2, "testdim", ds, origin_at=origin_at)

        if origin_at is None or origin_at == "lower":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(0.1, 1.9, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        elif origin_at == "centre":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(-0.9, 0.9, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        elif origin_at == "upper":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(-1.9, -0.1, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        else:
            assert False

    @pytest.mark.parametrize(
        "origin_at", [None, "lower", "centre", "upper", "expectfail"]
    )
    def test__1d_coord_from_spacing_da(self, origin_at):
        da = xr.DataArray(0.2 * np.ones(10), dims="testdim")
        da.attrs["metadata"] = {
            "bout_xdim": "x",
            "bout_ydim": "y",
            "bout_zdim": "z",
            "keep_xboundaries": True,
            "keep_yboundaries": True,
        }

        if origin_at == "expectfail":
            with pytest.raises(ValueError):
                coord = _1d_coord_from_spacing(da, "testdim", origin_at=origin_at)
            return

        coord = _1d_coord_from_spacing(da, "testdim", origin_at=origin_at)

        if origin_at is None or origin_at == "lower":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(0.1, 1.9, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        elif origin_at == "centre":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(-0.9, 0.9, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        elif origin_at == "upper":
            xrt.assert_allclose(
                coord,
                xr.Variable("testdim", np.linspace(-1.9, -0.1, 10)),
                rtol=2.0e-15,
                atol=0.0,
            )
        else:
            assert False
